# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.4.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if getattr(globals().get("__spec__"), "parent", None) or __package__ or "." in __name__:
    from . import _icalmapi
else:
    import _icalmapi

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


VEVENT = _icalmapi.VEVENT
VTODO = _icalmapi.VTODO
VJOURNAL = _icalmapi.VJOURNAL
IC2M_NO_RECIPIENTS = _icalmapi.IC2M_NO_RECIPIENTS
IC2M_APPEND_ONLY = _icalmapi.IC2M_APPEND_ONLY
IC2M_NO_ORGANIZER = _icalmapi.IC2M_NO_ORGANIZER
IC2M_NO_BODY = _icalmapi.IC2M_NO_BODY
class ICalToMapi(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _icalmapi.delete_ICalToMapi

    def ParseICal2(self, ical, charset, server_tz, arg5, flags):
        return _icalmapi.ICalToMapi_ParseICal2(self, ical, charset, server_tz, arg5, flags)

    def ParseICal(self, ical, charset, server_tz, mu, flags):
        return _icalmapi.ICalToMapi_ParseICal(self, ical, charset, server_tz, mu, flags)

    def GetItemCount(self):
        return _icalmapi.ICalToMapi_GetItemCount(self)

    def GetItemInfo(self, ulPosition):
        return _icalmapi.ICalToMapi_GetItemInfo(self, ulPosition)

    def GetItem(self, ulPosition, ulFlags, lpMessage):
        return _icalmapi.ICalToMapi_GetItem(self, ulPosition, ulFlags, lpMessage)

    def GetFreeBusyInfo(self, users):
        return _icalmapi.ICalToMapi_GetFreeBusyInfo(self, users)

    def GetNumInvalidProperties(self):
        return _icalmapi.ICalToMapi_GetNumInvalidProperties(self)

    def GetNumInvalidComponents(self):
        return _icalmapi.ICalToMapi_GetNumInvalidComponents(self)

# Register ICalToMapi in _icalmapi:
_icalmapi.ICalToMapi_swigregister(ICalToMapi)

def CreateICalToMapi(propobj, arg2, no_recipients):
    return _icalmapi.CreateICalToMapi(propobj, arg2, no_recipients)
M2IC_CENSOR_PRIVATE = _icalmapi.M2IC_CENSOR_PRIVATE
M2IC_NO_VTIMEZONE = _icalmapi.M2IC_NO_VTIMEZONE
class MapiToICal(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _icalmapi.delete_MapiToICal

    def AddMessage(self, lpMessage, strSrvTZ, ulFlags):
        return _icalmapi.MapiToICal_AddMessage(self, lpMessage, strSrvTZ, ulFlags)

    def AddBlocks(self, lpsFBblk, ulBlocks, tStart, tEnd, strOrganiser, strUser, strUID):
        return _icalmapi.MapiToICal_AddBlocks(self, lpsFBblk, ulBlocks, tStart, tEnd, strOrganiser, strUser, strUID)

    def Finalize(self, ulFlags):
        return _icalmapi.MapiToICal_Finalize(self, ulFlags)

    def ResetObject(self):
        return _icalmapi.MapiToICal_ResetObject(self)

# Register MapiToICal in _icalmapi:
_icalmapi.MapiToICal_swigregister(MapiToICal)

def CreateMapiToICal(arg1, charset):
    return _icalmapi.CreateMapiToICal(arg1, charset)
class mapitovcf(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _icalmapi.delete_mapitovcf

    def add_message(self, arg2):
        return _icalmapi.mapitovcf_add_message(self, arg2)

    def finalize(self):
        return _icalmapi.mapitovcf_finalize(self)

# Register mapitovcf in _icalmapi:
_icalmapi.mapitovcf_swigregister(mapitovcf)

def create_mapitovcf():
    return _icalmapi.create_mapitovcf()
class vcftomapi(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _icalmapi.delete_vcftomapi

    def parse_vcf(self, ical):
        return _icalmapi.vcftomapi_parse_vcf(self, ical)

    def get_item_count(self):
        return _icalmapi.vcftomapi_get_item_count(self)

    def get_item(self, arg2, section=0):
        return _icalmapi.vcftomapi_get_item(self, arg2, section)

# Register vcftomapi in _icalmapi:
_icalmapi.vcftomapi_swigregister(vcftomapi)

def create_vcftomapi(arg1):
    return _icalmapi.create_vcftomapi(arg1)




# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano.

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import sys

from MAPI import (
    MAPI_UNICODE, RELOP_EQ, TBL_BATCH, ECSTORE_TYPE_PUBLIC,
    WrapStoreEntryID
)
from MAPI.Defs import PpropFindProp
from MAPI.Struct import (
    MAPIErrorNotFound, SPropertyRestriction, SPropValue,
    MAPIErrorCollision
)
from MAPI.Tags import PR_EC_COMPANY_NAME_W, PR_EC_STOREGUID
from MAPI.Util import GetPublicStore

from .properties import Properties
from .quota import Quota
from .group import Group

from .defs import EID_EVERYONE
from .errors import (
    NotFoundError, DuplicateError
)
from .compat import (
    benc as _benc, bdec as _bdec,
)

try:
    from . import server as _server
except ImportError: # pragma: no cover
    _server = sys.modules[__package__ + '.server']
try:
    from . import store as _store
except ImportError: # pragma: no cover
    _store = sys.modules[__package__ + '.store']

class Company(Properties):
    """Company class

    For a multi-tenant Kopano setup, users are grouped in companies.
    Regular users are unaware of anything outside their own company.
    """

    def __init__(self, name, server=None):
        self.server = server or \
            _server.Server(_skip_check=True, parse_args=False)

        self._name = name = str(name)
        if name != 'Default': # TODO
            try:
                id_ = self.server.sa.ResolveCompanyName(name, MAPI_UNICODE)
                self._eccompany = self.server.sa.GetCompany(id_, MAPI_UNICODE)
            except MAPIErrorNotFound:
                raise NotFoundError("no such company: '%s'" % name)

        # TODO cached because GetPublicStore does not see changes..
        #     do we need folder & store notifications (slow)?
        self._public_store = None
        self._mapiobj = None

    @property
    def mapiobj(self):
        """Underlying MAPI object."""
        if not self._mapiobj:
            id_ = self._eccompany.CompanyID
            self._mapiobj = self.server.mapisession.OpenEntry(id_, None, 0)
        return self._mapiobj

    @property
    def companyid(self): # TODO single-tenant case
        """Company id."""
        if self._name != 'Default':
            return _benc(self._eccompany.CompanyID)

    @property
    def admin(self):
        """Company :class:`administrator <User>` in multi-tenant mode."""
        if self._name != 'Default':
            id_ = self._eccompany.AdministratorID
            ecuser = self.server.sa.GetUser(id_, MAPI_UNICODE)
            return self.server.user(ecuser.Username)

    @admin.setter
    def admin(self, user):
        self._eccompany.AdministratorID = user._ecuser.UserID
        self.server.sa.SetCompany(self._eccompany, MAPI_UNICODE)

    @property
    def hidden(self):
        """The company is hidden from the addressbook."""
        if self._name != 'Default':
            return bool(self._eccompany.IsHidden)
        return False

    @property
    def name(self):
        """Company name."""
        return self._name

    @name.setter
    def name(self, value):
        value = str(value)
        self._eccompany.Companyname = value
        self._name = value
        self.server.sa.SetCompany(self._eccompany, MAPI_UNICODE)

    def store(self, guid):
        """:class:`store <Store>` for the given GUID

        :param guid: store guid
        """
        if guid == 'public':
            if not self.public_store:
                raise NotFoundError(
                    "no public store for company '%s'" % self.name
                )
            return self.public_store
        else:
            return self.server.store(guid)

    def stores(self):
        """Return all company :class:`stores <Store>`."""
        if self.server.multitenant:
            table = self.server.sa.OpenUserStoresTable(MAPI_UNICODE)
            restriction = SPropertyRestriction(
                RELOP_EQ,
                PR_EC_COMPANY_NAME_W,
                SPropValue(PR_EC_COMPANY_NAME_W, self.name)
            )
            table.Restrict(restriction, TBL_BATCH)
            for row in table.QueryRows(2147483647, 0):
                prop = PpropFindProp(row, PR_EC_STOREGUID)
                if prop:
                    yield _store.Store(_benc(prop.Value), self.server)
            public_store = self.public_store # TODO why not included above?
            if public_store:
                yield public_store
        else:
            for store in self.server.stores():
                yield store

    @property
    def public_store(self):
        """Company public :class:`store <Store>`."""
        if not self._public_store:
            if self._name == 'Default': # TODO
                pubstore = GetPublicStore(self.server.mapisession)
                if pubstore is None:
                    self._public_store = None
                else:
                    self._public_store = \
                        _store.Store(mapiobj=pubstore, server=self.server)
            else:
                try:
                    entryid = self.server.ems.CreateStoreEntryID(None,
                        self._name, MAPI_UNICODE)
                except MAPIErrorNotFound:
                    self._public_store = None
                else:
                    self._public_store = \
                        _store.Store(entryid=_benc(entryid),server=self.server)

        return self._public_store

    def create_public_store(self):
        """Create company public :class:`store <Store>`."""
        if self._name == 'Default':
            try:
                storeid_rootid = self.server.sa.CreateStore(
                    ECSTORE_TYPE_PUBLIC, EID_EVERYONE)
            except MAPIErrorCollision:
                raise DuplicateError("public store already exists")
        else:
            try:
                storeid_rootid = self.server.sa.CreateStore(
                    ECSTORE_TYPE_PUBLIC, self._eccompany.CompanyID)
            except MAPIErrorCollision:
                raise DuplicateError(
                    "public store already exists for company '%s'" % self.name)

        store_entryid = WrapStoreEntryID(0, b'zarafa6client.dll',
            storeid_rootid[0][:-4]) + self.server.pseudo_url + b'\x00'

        self._public_store = _store.Store(entryid=_benc(store_entryid),
            server=self.server)
        return self._public_store

    def hook_public_store(self, store):
        """Hook company public :class:`store <Store>`.

        :param store: :class:`store <Store>` to hook as public store
        """
        if self._name == 'Default':
            try:
                self.server.sa.HookStore(ECSTORE_TYPE_PUBLIC, EID_EVERYONE,
                    _bdec(store.guid))
            except MAPIErrorCollision:
                raise DuplicateError("hooked public store already exists")
        else:
            try:
                self.server.sa.HookStore(ECSTORE_TYPE_PUBLIC,
                    _bdec(self.companyid), _bdec(store.guid))
            except MAPIErrorCollision:
                raise DuplicateError(
                    "hooked public store already exists for company '%s'" %
                    self.name)

        self._public_store = store

    def unhook_public_store(self):
        """Unhook company public :class:`store <Store>`."""
        if self._name == 'Default':
            try:
                self.server.sa.UnhookStore(ECSTORE_TYPE_PUBLIC, EID_EVERYONE)
            except MAPIErrorNotFound:
                raise NotFoundError("no hooked public store")
        else:
            try:
                self.server.sa.UnhookStore(ECSTORE_TYPE_PUBLIC,
                    _bdec(self.companyid))
            except MAPIErrorNotFound:
                raise NotFoundError(
                    "no hooked public store for company '%s'" % self.name)
        self._public_store = None

    def user(self, name, create=False):
        """Return :class:`user <User>` with given name.

        :param name: user name
        :param create: create user if it doesn't exist (default False)

        :raises: NotFoundError
        """
        if '@' not in name and self._name != 'Default':
            name = name + '@' + self._name
        try:
            return self.server.user(name)
        except NotFoundError:
            if create:
                return self.create_user(name)
            else:
                raise

    def get_user(self, name):
        """Return :class:`user <User>` with given name or *None* if not found.

        :param name: user name
        """
        try:
            return self.user(name)
        except NotFoundError:
            pass

    def users(self, **kwargs):
        """
        Return all :class:`users <User>` within company.

        :param parse: filter users on cli argument --user (default True)
        :param system: include system users (default False)
        """

        if 'remote' not in kwargs:
            kwargs['remote'] = True

        for user in self.server.users(_company=self, **kwargs):
            yield user

    def create_user(self, name, password=None):
        """Create a new :class:`user <User>` within the company.

        :param name: user name
        :param password: password (default None)

        :raises: DuplicateError
        """
        name = name.split('@')[0]
        if self._name == 'Default':
            self.server.create_user(name, password=password)
        else:
            self.server.create_user(name, password=password,
                company=self._name)
        return self.user(name)

    def group(self, name, create=False):
        """Return :class:`group <Group>` with given name.

        :param name: group name
        :param create: create group if it doesn't exist (default False)

        :raises: NotFoundError
        """
        for group in self.groups(): # TODO
            if group.name == name:
                return group
        if create:
            return self.create_group(name)
        else:
            raise NotFoundError("no such group: '%s'" % name)

    def get_group(self, name):
        """Return :class:`group <Group>` with given name or *None*
        if not found.

        :param name: group name
        """
        try:
            return self.group(name)
        except NotFoundError:
            pass

    def groups(self):
        """Return all :class:`groups <Group>` within the company."""
        if self.name == 'Default': # TODO
            for group in self.server.groups():
                yield group
        else:
            for ecgroup in self.server.sa.GetGroupList(
                    self._eccompany.CompanyID, MAPI_UNICODE):
                yield Group(ecgroup.Groupname, self.server)

    def create_group(self, name):
        """Create :class:`group <Group>`.

        :param name: group name
        :raises: DuplicateError
        """
        name = name.split('@')[0]
        if self._name == 'Default':
            self.server.create_group(name)
        else:
            self.server.create_group(name+'@'+self._name)
        return self.group(name)

    def admins(self):
        """Return all company :class:`admins <User>`."""
        for ecuser in self.server.sa.GetRemoteAdminList(
                self._eccompany.CompanyID, MAPI_UNICODE):
            yield self.server.user(ecuser.Username)

    def add_admin(self, user):
        """Add user to admin list

        :param user: the user class:`user <User>`
        :raises: DuplicateError
        """
        try:
            self.server.sa.AddUserToRemoteAdminList(user._ecuser.UserID,
                self._eccompany.CompanyID)
        except MAPIErrorCollision:
            raise DuplicateError(
                "user '%s' already admin for company '%s'" %
                (user.name, self.name))

    def remove_admin(self, user):
        """Remove user from admin list

        :param user: the :class:`user <User>`
        :raises: NotFoundError
        """
        try:
            self.server.sa.DelUserFromRemoteAdminList(user._ecuser.UserID,
                self._eccompany.CompanyID)
        except MAPIErrorNotFound:
            raise NotFoundError("user '%s' no admin for company '%s'" %
                (user.name, self.name))

    def views(self):
        """Return all :class:`remote viewers <Company>` (other companies
        to which the address book is visible)."""
        for eccompany in self.server.sa.GetRemoteViewList(
                self._eccompany.CompanyID, MAPI_UNICODE):
            yield self.server.company(eccompany.Companyname)

    def add_view(self, company):
        """Add :class:`remote viewer <Company>`."""
        try:
            self.server.sa.AddCompanyToRemoteViewList(
                company._eccompany.CompanyID, self._eccompany.CompanyID)
        except MAPIErrorCollision:
            raise DuplicateError(
                "company '%s' already in view-list for company '%s'" %
                (company.name, self.name))

    def remove_view(self, company):
        """Remove :class:`remote viewer <Company>`."""
        try:
            self.server.sa.DelCompanyFromRemoteViewList(
                company._eccompany.CompanyID, self._eccompany.CompanyID)
        except MAPIErrorNotFound:
            raise NotFoundError(
                "company '%s' not in view-list for company '%s'" %
                (company.name, self.name))

    @property
    def quota(self):
        """Company :class:`Quota`."""
        if self._name == 'Default':
            return Quota(self.server, None)
        else:
            return Quota(self.server, self._eccompany.CompanyID)

    def __eq__(self, c):
        if isinstance(c, Company):
            return self.companyid == c.companyid
        return False

    def __ne__(self, c):
        return not self == c

    def __contains__(self, u):
        return u in self.users()

    def __unicode__(self):
        return u"Company('%s')" % self._name



# -*- indent-tabs-mode: nil -*-
# SPDX-License-Identifier: AGPL-3.0-only

import inspect
import sys

import MAPICore
from MAPI.Defs import *

def _convert(s):
    if isinstance(s, bytes):
        return s.decode('ascii')
    else:
        return s

class MAPIStruct:
    def __init__(self): pass
    def __eq__(self, other):
        if other is None:
            return False
        return self.__dict__ == other.__dict__
    def __repr__(self):
        return repr(self.__dict__)
    def __setstate__(self, d):
        # XXX pickle with python2, unpickle with python3 (encoding='bytes')
        for k, v in d.items():
            setattr(self, _convert(k), v)

class SPropValue(MAPIStruct):
    def __init__(self, ulPropTag, Value):
        self.ulPropTag = ulPropTag
        self.Value = Value

    def __repr__(self):
        return "SPropValue(0x%08X, %s)" % (self.ulPropTag, repr(self.Value))

    def __eq__(self, other):
        return self.ulPropTag == other.ulPropTag and self.Value == other.Value

    def __lt__(self, other):
        if self.ulPropTag != other.ulPropTag:
            return self.ulPropTag < other.ulPropTag
        return self.Value < other.Value

    def __gt__(self, other):
        if self.ulPropTag != other.ulPropTag:
            return self.ulPropTag > other.ulPropTag
        return self.Value > self.Value

    def __le__(self, other):
        if self.ulPropTag != other.ulPropTag:
            return self.ulPropTag <= other.ulPropTag
        return self.Value <= other.Value

    def __ge__(self, other):
        if self.ulPropTag != other.ulPropTag:
            return self.ulPropTag >= other.ulPropTag
        return self.Value >= self.Value

    def __hash__(self):
        return hash(self.ulPropTag) + hash(self.Value)

class SSort(MAPIStruct):
    def __init__(self, ulPropTag, ulOrder):
        self.ulPropTag = ulPropTag
        self.ulOrder = ulOrder

    def __repr__(self):
        sorts = [ 'TABLE_SORT_ASCEND', 'TABLE_SORT_DESCEND' ]
        return "SSort(%#x, %s)" % (self.ulPropTag, sorts[self.ulOrder])

class SSortOrderSet(MAPIStruct):
    def __init__(self, sorts, cCategories, cExpanded):
        self.aSort = sorts
        self.cCategories = cCategories
        self.cExpanded = cExpanded

    def __repr__(self):
        return "SSortOrderSet(%r, %r, %r)" % (self.aSort, self.cCategories, self.cExpanded)

class MAPINAMEID(MAPIStruct):
    def __init__(self, guid, kind, id):
        self.guid = guid
        self.kind = kind
        self.id = id

    def __hash__(self):
        return (self.guid, self.kind, self.id).__hash__()

    def __repr__(self):
        if (self.kind == MAPICore.MNID_ID):
            return "MAPINAMEID(%r, MNID_ID, %r)" % (self.guid, self.id)
        else:
            return "MAPINAMEID(%r, MNID_STRING, %r)" % (self.guid, self.id)

class SPropProblem(MAPIStruct):
    def __init__(self, index, tag, scode):
        self.ulIndex = index
        self.ulPropTag = tag
        self.scode = scode

class SAndRestriction(MAPIStruct):
    def __init__(self, sub):
        self.rt = MAPICore.RES_AND
        self.lpRes = sub
    def __repr__(self):
        return 'SAndRestriction(%r)' % (self.lpRes)

class SOrRestriction(MAPIStruct):
    def __init__(self, sub):
        self.rt = MAPICore.RES_OR
        self.lpRes = sub
    def __repr__(self):
        return 'SOrRestriction(%r)' % (self.lpRes)

class SNotRestriction(MAPIStruct):
    def __init__(self, sub):
        self.rt = MAPICore.RES_NOT
        self.lpRes = sub
    def __repr__(self):
        return 'SNotRestriction(%r)' % (self.lpRes)

class SContentRestriction(MAPIStruct):
    def __init__(self, fuzzy, proptag, prop):
        self.rt = MAPICore.RES_CONTENT
        self.ulFuzzyLevel = fuzzy
        self.ulPropTag = proptag
        self.lpProp = prop
    def __repr__(self):
        return 'SContentRestriction(%r,%#x,%r)' % (self.ulFuzzyLevel, self.ulPropTag, self.lpProp)

class SBitMaskRestriction(MAPIStruct):
    def __init__(self, relBMR, ulPropTag, ulMask):
        self.rt = MAPICore.RES_BITMASK
        self.relBMR = relBMR
        self.ulPropTag = ulPropTag
        self.ulMask = ulMask
    def __repr__(self):
        return 'SBitMaskRestriction(%r,%#x,%r)' % (self.relBMR, self.ulPropTag, self.ulMask)

class SPropertyRestriction(MAPIStruct):
    def __init__(self, relop, ulPropTag, prop):
        self.rt = MAPICore.RES_PROPERTY
        self.ulPropTag = ulPropTag
        self.relop = relop
        self.lpProp = prop
    def __repr__(self):
        return 'SPropertyRestriction(%r,%#x,%r)' %(self.relop, self.ulPropTag, self.lpProp)

class SComparePropsRestriction(MAPIStruct):
    def __init__(self, relop, ulPropTag1, ulPropTag2):
        self.rt = MAPICore.RES_COMPAREPROPS
        self.relop = relop
        self.ulPropTag1 = ulPropTag1
        self.ulPropTag2 = ulPropTag2
    def __repr__(self):
        return 'SComparePropsRestriction(%r,%#x,%#x)' % (self.relop, self.ulPropTag1, self.ulPropTag2)

class SSizeRestriction(MAPIStruct):
    def __init__(self, relop, ulPropTag, cb):
        self.rt = MAPICore.RES_SIZE
        self.relop = relop
        self.ulPropTag = ulPropTag
        self.cb = cb
    def __repr__(self):
        return 'SSizeRestriction(%r,%#x,%r)' % (self.relop, self.ulPropTag, self.cb)

class SExistRestriction(MAPIStruct):
    def __init__(self, proptag):
        self.rt = MAPICore.RES_EXIST
        self.ulPropTag = proptag
    def __repr__(self):
        return 'SExistRestriction(%#x)' % (self.ulPropTag)

class SSubRestriction(MAPIStruct):
    def __init__(self, ulSubObject, res):
        self.rt = MAPICore.RES_SUBRESTRICTION
        self.ulSubObject = ulSubObject
        self.lpRes = res
    def __repr__(self):
        return 'SSubRestriction(%#x,%r)' % (self.ulSubObject, self.lpRes)

class SCommentRestriction(MAPIStruct):
    def __init__(self, res, prop):
        self.rt = MAPICore.RES_COMMENT
        self.lpRes = res
        self.lpProp = prop
    def __repr__(self):
        return 'SCommentRestriction(%r,%r)' %(self.lpRes, self.lpProp)

class actMoveCopy(MAPIStruct):
    def __init__(self, store, folder):
        self.StoreEntryId = store
        self.FldEntryId = folder
    def __repr__(self):
        return 'actMoveCopy(%r,%r)' % (self.StoreEntryId, self.FldEntryId)

class actReply(MAPIStruct):
    def __init__(self, entryid, guid):
        self.EntryId = entryid
        self.guidReplyTemplate = guid
    def __repr__(self):
        return 'actReply(%r,%r)' % (self.EntryId, self.guidReplyTemplate)

class actDeferAction(MAPIStruct):
    def __init__(self, deferMsg):
        self.data = deferMsg
    def __repr__(self):
        return 'actDeferMsg(%r)' % (self.data)

class actBounce(MAPIStruct):
    def __init__(self, code):
        self.scBounceCode = code
    def __repr__(self):
        return 'actBounce(%r)' % (self.scBounceCode)

class actFwdDelegate(MAPIStruct):
    def __init__(self, adrlist):
        self.lpadrlist = adrlist
    def __repr__(self):
        return 'actFwdDelegate(%r)' % (self.lpadrlist)

class actTag(MAPIStruct):
    def __init__(self, tag):
        self.propTag = tag
    def __repr__(self):
        return 'actTag(%r)' % (self.propTag)

class ACTION(MAPIStruct):
    def __init__(self, acttype, flavor, res, proptagarray, flags, actobj):
        self.acttype = acttype
        self.ulActionFlavor = flavor
        self.lpRes = res
        self.lpPropTagArray = proptagarray
        self.ulFlags = flags
        # any of the above action
        self.actobj = actobj
    def __repr__(self):
        return 'ACTION(%r,%r,%r,%r,%#x,%r)' % (self.acttype,self.ulActionFlavor,self.lpRes,self.lpPropTagArray,self.ulFlags,self.actobj)

class ACTIONS(MAPIStruct):
    def __init__(self, version, actions):
        self.ulVersion = version
        self.lpAction = actions
    def __repr__(self):
        return 'ACTIONS(%r,%r)' % (self.ulVersion, self.lpAction)

# pylint (let's replace the horror which comes after this with a simple list)
MAPIErrorNotFound = None
MAPIErrorInterfaceNotSupported = None
MAPIErrorNotEnoughMemory = None
MAPIErrorNoSupport = None
MAPIErrorInvalidParameter = None
MAPIErrorNoAccess = None
MAPIErrorInvalidEntryid = None
MAPIErrorUnknownEntryid = None
MAPIErrorNoSupport = None
MAPIErrorNetworkError = None
MAPIErrorDiskError = None
MAPIErrorLogonFailed = None

class MAPIError(Exception):
    _errormap = {}

    @staticmethod
    def _initialize_errors():
        error_prefix = 'MAPI_E_'
        for name, value in inspect.getmembers(sys.modules['MAPICore']):
            if name.startswith(error_prefix):
                clsname = 'MAPIError' + ''.join(s.capitalize() for s in name[len(error_prefix):].split('_'))
                def construct_class(hr, descr):
                    if issubclass(MAPIError, object):
                        return type(clsname, (MAPIError,), {'__init__': lambda self: MAPIError.__init__(self, hr, descr)})
                    else:
                        class MAPIErrorDynamic(MAPIError):
                            def __init__(self):
                                MAPIError.__init__(self, hr, descr)
                        return MAPIErrorDynamic
                t = construct_class(value, name)
                setattr(sys.modules[__name__], clsname, t)
                MAPIError._errormap[value] = t

    @staticmethod
    def _get_type(hr):
        '''Returns the type of the MAPIError subclass that represents the given error code.'''
        return MAPIError._errormap.get(hr)

    @staticmethod
    def from_hresult(hr):
        '''Use MAPIError.from_hresult(hr) to create a MAPIError subclass based on the error
        code. If no subclass is defined for the error code, a MAPIError instance is
        returned.
        '''
        t = MAPIError._get_type(hr)
        if t:
            return t()
        return MAPIError(hr)

    def __init__(self, hr, descr=None):
        self.hr = hr
        self.descr = descr

    def __repr__(self):
        return "MAPI error %X (%s)" % (self.hr, self.descr)

    def __str__(self):
        return self.__repr__()

class READSTATE(MAPIStruct):
    def __init__(self, SourceKey, ulFlags):
        self.SourceKey = SourceKey
        self.ulFlags = ulFlags

class MVPROPMAP(MAPIStruct):
    def __init__(self, ulPropId, Values):
        self.ulPropId = ulPropId
        self.Values = Values

    def __str__(self):
        return 'MVPROPMAP(%s)' % (self.ulPropId)

class ECUSER(MAPIStruct):
    def __init__(self, Username, Password, Email, FullName, Servername = None, Class = 0x10001, IsAdmin = False, IsHidden = False, Capacity = 0, UserID = None, MVPropMap = None):
        self.Username = Username
        self.Password = Password
        self.Email = Email
        self.FullName = FullName
        self.Servername = Servername
        self.Class = Class
        self.IsAdmin = IsAdmin
        self.IsHidden = IsHidden
        self.Capacity = Capacity
        self.UserID = UserID
        self.MVPropMap = MVPropMap
    def __str__(self):
        return 'ECUSER(%s,%s,%s)' % (self.Username, self.Email, self.FullName)

# @todo propmap?
class ECGROUP(MAPIStruct):
    def __init__(self, Groupname, Fullname, Email, IsHidden = False, GroupID = None, MVPropMap = None):
        self.Groupname = Groupname
        self.Fullname = Fullname
        self.Email = Email
        self.IsHidden = IsHidden
        self.GroupID = GroupID
        self.MVPropMap = MVPropMap

# @todo propmap?
class ECCOMPANY(MAPIStruct):
    def __init__(self, Companyname, Servername, IsHidden = False, CompanyID = None, MVPropMap = None, AdministratorID = None):
        self.Companyname = Companyname
        self.Servername = Servername
        self.IsHidden = IsHidden
        self.CompanyID = CompanyID
        self.MVPropMap = MVPropMap
        self.AdministratorID = AdministratorID

class ECSERVER(MAPIStruct):
    def __init__(self, Name, FilePath, HttpPath, SslPath, PreferedPath, Flags):
        self.Name = Name
        self.FilePath = FilePath
        self.HttpPath = HttpPath
        self.SslPath = SslPath
        self.PreferedPath = PreferedPath
        self.Flags = Flags

class ERROR_NOTIFICATION(MAPIStruct):
    def __init__(self, lpEntryID, scode, ulFlags, lpMAPIError):
        self.lpEntryID = lpEntryID
        self.scode = scode
        self.ulFlags = ulFlags
        self.lpMAPIError = lpMAPIError

class NEWMAIL_NOTIFICATION(MAPIStruct):
    def __init__(self, lpEntryID, lpParentID, ulFlags, lpszMessageClass, ulMessageFlags):
        self.lpEntryID = lpEntryID
        self.lpParentID = lpParentID
        self.ulFlags = ulFlags
        self.lpszMessageClass = lpszMessageClass
        self.ulMessageFlags = ulMessageFlags

class OBJECT_NOTIFICATION(MAPIStruct):
    def __init__(self, ulEventType, lpEntryID, ulObjType, lpParentID, lpOldID, lpOldParentID, lpPropTagArray):
        self.ulEventType = ulEventType
        self.lpEntryID = lpEntryID
        self.ulObjType = ulObjType
        self.lpParentID = lpParentID
        self.lpOldID = lpOldID
        self.lpOldParentID = lpOldParentID
        self.lpPropTagArray = lpPropTagArray

class TABLE_NOTIFICATION(MAPIStruct):
    def __init__(self, ulTableEvent, hResult, propIndex, propPrior, row):
        self.ulTableEvent = ulTableEvent
        self.hResult = hResult
        self.propIndex = propIndex
        self.propPrior = propPrior
        self.row = row

class ROWENTRY(MAPIStruct):
    def __init__(self, ulRowFlags, rgPropVals):
        self.ulRowFlags = ulRowFlags
        self.rgPropVals = rgPropVals

class ECQUOTA(MAPIStruct):
    def __init__(self, bUseDefaultQuota, bIsUserDefaultQuota, llWarnSize, llSoftSize, llHardSize):
        self.bUseDefaultQuota = bUseDefaultQuota
        self.bIsUserDefaultQuota = bIsUserDefaultQuota
        self.llWarnSize = llWarnSize
        self.llSoftSize = llSoftSize
        self.llHardSize = llHardSize

class ECQUOTASTATUS(MAPIStruct):
    def __init__(self, StoreSize, QuotaStatus):
        self.StoreSize = StoreSize
        self.QuotaStatus = QuotaStatus

class STATSTG(MAPIStruct):
    def __init__(self, cbSize):
        self.cbSize = cbSize

class SYSTEMTIME(MAPIStruct):
    def __init__(self, wYear, wMonth, wDayOfWeek, wDay, wHour, wMinute, wSecond, wMilliseconds):
        self.wYear = wYear
        self.wMonth = wMonth
        self.wDayOfWeek = wDayOfWeek
        self.wDay = wDay
        self.wHour = wHour
        self.wMinute = wMinute
        self.wSecond = wSecond
        self.wMilliseconds = wMilliseconds

class FreeBusyBlock(MAPIStruct):
    def __init__(self, start, end, status):
        self.start = start
        self.end = end
        self.status = status


MAPIError._initialize_errors()



# SPDX-License-Identifier: AGPL-3.0-only
"""
High-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)

Some goals:

- To be fully object-oriented, pythonic, layer above MAPI
- To be usable for many common system administration tasks
- To provide full access to the underlying MAPI layer if needed
- To return/accept binary identifiers in readable form
- To raise well-described exceptions if something goes wrong

"""

import sys
import types

from .version import __version__
from .config import Config, CONFIG
from .errors import (
    Error, ConfigError, DuplicateError, NotFoundError, LogonError,
    NotSupportedError, ArgumentError,
)
from .server import Server as _Server
from .address import Address
from .appointment import Appointment
from .attachment import Attachment
from .attendee import Attendee
from .autoaccept import AutoAccept
from .autoprocess import AutoProcess
from .company import Company
from .compat import (
    set_bin_encoding, set_missing_none, hex, benc, bdec
)
from .contact import Contact
from .delegation import Delegation
from .distlist import DistList
from .folder import Folder
from .freebusy import FreeBusyBlock, FreeBusy
from .group import Group
from .item import Item
from .log import log_exc, QueueListener, logger
from .meetingrequest import MeetingRequest
from .notification import Notification
from .outofoffice import OutOfOffice
from .property_ import Property
from .properties import Properties
from .permission import Permission
from .picture import Picture
from .quota import Quota
from .recurrence import Recurrence, Occurrence
from .restriction import Restriction
from .rule import Rule
from .store import Store
from .table import Table
from .user import User
from .parser import parser
from .service import Service, Worker, server_socket, client_socket


# interactive shortcuts

# TODO add kopano.servers?

class Module(types.ModuleType):
    def __init__(self):
        super().__init__(__name__)
        self.__dict__.update(sys.modules[__name__].__dict__)

        class DeprecatedServer(_Server):
            def __init__(self, *args, **kwargs):
                kwargs['_skip_check'] = False  # force deprecation warning
                kwargs['parse_args'] = kwargs.get('parse_args', True)
                super().__init__(*args, **kwargs)

        class CallableSubModule(types.ModuleType):
            def __init__(self, name, callFunc):
                super().__init__(__name__ + '.' + name)
                self.__dict__.update(sys.modules[__name__ + '.' + name].__dict__)
                self.__callFunc = callFunc

            def __call__(self, *args, **kwargs):
                return self.__callFunc(*args, **kwargs)

        # Make our submodules which clash with our public API available while
        # retaining callable public API functions.
        self.server = CallableSubModule('server', self.__serverFactory)
        self.user = CallableSubModule('user', self.__userFactory)
        self.group = CallableSubModule('group', self.__groupFactory)
        self.store = CallableSubModule('store', self.__storeFactory)
        self.company = CallableSubModule('company', self.__companyFactory)

        # Deprecated API.
        self.Server = DeprecatedServer

        # Internals.
        self.__server = None

    @property
    def _server(self):
        if not self.__server:
            self.__server = _Server(_skip_check=True, parse_args=False)
        return self.__server

    # TODO add 'name' argument to lookup node?
    def __serverFactory(self, *args, **kwargs):
        kwargs['parse_args'] = kwargs.get('parse_args', False)
        return _Server(*args, **kwargs)

    # TODO add servers() for multiserver

    def __userFactory(self, *args, **kwargs):
        return self._server.user(*args, **kwargs)

    def users(self, *args, **kwargs):
        return self._server.users(*args, **kwargs)

    def __groupFactory(self, *args, **kwargs):
        return self._server.group(*args, **kwargs)

    def groups(self, *args, **kwargs):
        return self._server.groups(*args, **kwargs)

    def __storeFactory(self, *args, **kwargs):
        return self._server.store(*args, **kwargs)

    def stores(self, *args, **kwargs):
        return self._server.stores(*args, **kwargs)

    def __companyFactory(self, *args, **kwargs):
        return self._server.company(*args, **kwargs)

    def companies(self, *args, **kwargs):
        return self._server.companies(*args, **kwargs)

    @property
    def public_store(self):
        return self._server.public_store


sys.modules[__name__] = Module()



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano.

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import atexit
import codecs
import gc
import datetime
import functools
import os
import time
import socket
import sys
import warnings

from urllib.parse import urlparse

from MAPI import (
    MAPI_UNICODE, MDB_WRITE, RELOP_EQ, MAPI_MAILUSER, TBL_BATCH,
    ECSTORE_TYPE_PRIVATE, DT_REMOTE_MAILUSER, RELOP_NE, FL_FULLSTRING,
    FL_IGNORECASE, BOOKMARK_BEGINNING, EC_OVERRIDE_HOMESERVER, WrapStoreEntryID
)
from MAPI.Util import (
    GetDefaultStore, OpenECSession
)
from MAPI.Defs import (
    HrGetOneProp,
)

from MAPI.Struct import (
    SPropertyRestriction, SPropValue, ECCOMPANY, ECGROUP, ECUSER,
    MAPIErrorNotFound, MAPIErrorNoSupport, MAPIErrorCollision,
    MAPIErrorLogonFailed, MAPIErrorNetworkError, MAPIErrorDiskError,
    SAndRestriction, SContentRestriction,
)
from MAPI.Tags import (
    PR_ACCOUNT_W, PURGE_CACHE_ALL, PR_DISPLAY_NAME_W, PR_ENTRYID,
    PR_STORE_RECORD_KEY, PR_MAPPING_SIGNATURE,
    PR_EC_STATSTABLE_SYSTEM, PR_EC_STATSTABLE_SESSIONS,
    PR_EC_STATSTABLE_USERS, PR_EC_STATSTABLE_COMPANY, PR_DISPLAY_NAME,
    PR_EC_STATSTABLE_SERVERS, PR_EC_STATS_SERVER_HTTPSURL,
    PR_STORE_ENTRYID, EC_PROFILE_FLAGS_NO_UID_AUTH, PR_EC_STATS_SYSTEM_VALUE,
    EC_PROFILE_FLAGS_NO_NOTIFICATIONS, EC_PROFILE_FLAGS_OIDC,
    PR_FREEBUSY_ENTRYIDS,
    PR_OBJECT_TYPE, PR_DISPLAY_TYPE,
)
from MAPI.Tags import (
    IID_IMsgStore, IID_IMAPITable, IID_IExchangeManageStore,
    IID_IECServiceAdmin
)

from .errors import (
    Error, NotFoundError, DuplicateError, NotSupportedError,
    LogonError, ArgumentError, _DeprecationWarning
)
from .log import LOG, _loglevel

from .parser import parser
from .table import Table
from .company import Company
from .group import Group
from .query import _query_to_restriction

from .compat import (
    benc as _benc,
    bdec as _bdec
)

try:
    from . import user as _user
except ImportError: # pragma: no cover
    _user = sys.modules[__package__ + '.user']
try:
    from . import config as _config
except ImportError: # pragma: no cover
    _config = sys.modules[__package__ + '.config']
from . import ics as _ics
try:
    from . import store as _store
except ImportError: # pragma: no cover
    _store = sys.modules[__package__ + '.store']
try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

def _timed_cache(seconds=0, minutes=0, hours=0, days=0):
    # used with permission from will mcgugan, https://www.willmcgugan.com
    time_delta = datetime.timedelta(
        seconds=seconds, minutes=minutes, hours=hours, days=days)

    def decorate(f):
        f._updates = {}
        f._results = {}

        def do_cache(*args, **kwargs):
            key = tuple(sorted(kwargs.items()))

            updates = f._updates
            results = f._results

            t = datetime.datetime.now()
            updated = updates.get(key, t)

            if key not in results or t - updated > time_delta:
                # calculate
                updates[key] = t
                result = f(*args, **kwargs)
                results[key] = result
                return result
            else:
                # cache
                return results[key]
        return do_cache
    return decorate

def instance_method_lru_cache(*cache_args, **cache_kwargs):
    # Just like functools.lru_cache, but a new cache is created for each
    # instance of the class that owns the method this is applied to.
    # Snippet by Alex Fraser (https://github.com/z0u, alex@phatcore.com)
    def cache_decorator(func):
        @functools.wraps(func)
        def cache_factory(self, *args, **kwargs):
            # Wrap the function in a cache by calling the decorator
            instance_cache = functools.lru_cache(
                *cache_args, **cache_kwargs)(func)
            # Bind the decorated function to the instance to make it a method
            instance_cache = instance_cache.__get__(self, self.__class__)
            setattr(self, func.__name__, instance_cache)
            # Call the instance cache now. Next time the method is called, the
            # call will go directly to the instance cache and not via this
            # decorator.
            return instance_cache(*args, **kwargs)
        return cache_factory
    return cache_decorator

# python does not always seem to release modules at shutdown:
#
# https://bugs.python.org/issue9072
#
# also, for the tests we've seen SWIG objects end up here:
#
# https://docs.python.org/2/library/gc.html#gc.garbage
#
# in both cases, store objects are not finalized, so the associated memory
# on the C side is not released, causing valgrind to report many leaks.
#
# to avoid valgrind going nuts, we manually clean up references at shutdown.
# for the second case, this may only be a work-around as it may be too late.
#
# for python3, both problems may have been solved already:
#
# https://www.python.org/dev/peps/pep-0442/
#
# and if not, python3 provides a work-around we could use, so manual
# clearing can be done each collection:
#
# https://docs.python.org/3/library/gc.html#gc.callbacks
#
# of course we might also try to prevent cycles in the first place.
#
# TODO valgrind python3

def _cleanup_mapistores():
    for obj in gc.get_objects():
        if isinstance(obj, (Server, _store.Store)):
            obj.__dict__.clear()
        elif (hasattr(obj, '__class__') and \
              obj.__class__.__name__ in ('IMAPISession', 'IMsgStore',
                  'IECServiceAdmin', 'IExchangeManageStore')):
            obj.__dict__.clear()

    del gc.garbage[:]
    gc.collect()

if os.getenv('ZCPSRCDIR'):
    atexit.register(_cleanup_mapistores)

class Options(object):
    pass

class Server(object):
    """Server class.

    Abstraction for Kopano servers. A MAPI session is automatically setup,
    according to the passed arguments and environment.
    """

    def __init__(self, options=None, config=None, sslkey_file=None,
            sslkey_pass=None, server_socket=None, auth_user=None,
            auth_pass=None, log=None, service=None, mapisession=None,
            parse_args=False, notifications=False, store_cache=True,
            oidc=False, _skip_check=True):
        """
        Create Server instance.

        By default, tries to connect to a storage server as configured in
        ``/etc/kopano/admin.cfg`` or at UNIX socket
        ``/var/run/kopano/server.sock``

        Looks at command-line to see if another server address or other
        related options were given (such as -c, -s, -k, -p)

        :param server_socket: similar to 'server_socket' option in config file
        :param sslkey_file: similar to 'sslkey_file' option in config file
        :param sslkey_pass: similar to 'sslkey_pass' option in config file
        :param config: path of configuration file containing common server
            options, for example ``/etc/kopano/admin.cfg``
        :param auth_user: username to user for user authentication
        :param auth_pass: password to use for user authentication
        :param options: OptionParser instance to get settings from
            (see:func:`parser`)
        :param parse_args: set this True if cli arguments should be parsed
        """
        self.options = options
        self.config = config
        self.sslkey_file = sslkey_file
        self.sslkey_pass = sslkey_pass
        self.server_socket = server_socket
        self.service = service
        if log: # TODO deprecate?
            self.log = log
        elif service:
            self.log = service.log
        elif config:
            self.log = LOG
            self.log.setLevel(_loglevel(options, config))
        else:
            self.log = LOG
        self.mapisession = mapisession
        self.store_cache = store_cache

        if not _skip_check:
            warnings.warn('use kopano.server instead of kopano.Server',
                _DeprecationWarning)

        if not self.mapisession:
            # get cmd-line options
            if not self.options:
                if parse_args:
                    self.options, _ = parser().parse_args()
                else:
                    self.options = Options()

            # determine config file
            if config:
                pass
            elif getattr(self.options, 'config_file', None):
                config_file = os.path.abspath(self.options.config_file)
                config = _config.Config(
                    None, filename=self.options.config_file)
            else:
                config_file = '/etc/kopano/admin.cfg'
                try:
                    open(config_file) # check if accessible
                    config = _config.Config(None, filename=config_file)
                except IOError:
                    pass
            self.config = config

            # get defaults
            if os.getenv("KOPANO_SOCKET"): # env variable used in testset
                self.server_socket = os.getenv("KOPANO_SOCKET")
            elif config:
                if (not (server_socket or \
                    getattr(self.options, 'server_socket', None))):
                    self.server_socket = config.get('server_socket')
                    self.sslkey_file = config.get('sslkey_file')
                    self.sslkey_pass = config.get('sslkey_pass')
            self.server_socket = self.server_socket or "default:"

            # override with explicit or command-line args
            self.server_socket = server_socket or \
                getattr(self.options, 'server_socket', None) or \
                self.server_socket
            self.sslkey_file = sslkey_file or \
                getattr(self.options, 'sslkey_file', None) or \
                self.sslkey_file
            self.sslkey_pass = sslkey_pass or \
                getattr(self.options, 'sslkey_pass', None) or \
                self.sslkey_pass

            # make actual connection. in case of service,
            # wait until this succeeds.
            self.auth_user = auth_user or \
                getattr(self.options, 'auth_user', None) or \
                ''
            self.auth_pass = auth_pass or \
                getattr(self.options, 'auth_pass', None) or \
                ''

            flags = 0
            self.notifications = notifications
            if not notifications:
                flags |= EC_PROFILE_FLAGS_NO_NOTIFICATIONS

            # Username and password was supplied, so let us do verification
            # (OpenECSession will not check password unless this parameter
            # is provided)
            if self.auth_user and self.auth_pass:
                flags |= EC_PROFILE_FLAGS_NO_UID_AUTH

            if oidc:
                flags |= EC_PROFILE_FLAGS_OIDC
            elif not self.auth_user:
                self.auth_user = "SYSTEM"

            while True:
                try:
                    self.mapisession = OpenECSession(
                        self.auth_user,
                        self.auth_pass,
                        self.server_socket,
                        sslkey_file=self.sslkey_file,
                        sslkey_pass=self.sslkey_pass,
                        flags=flags)
                    break
                except (MAPIErrorNetworkError, MAPIErrorDiskError):
                    if service:
                        self.log.warning("could not connect to server at \
'%s', retrying in 5 sec", self.server_socket)
                        time.sleep(5)
                    else:
                        raise Error("could not connect to server at '%s'" %
                            self.server_socket)
                except MAPIErrorLogonFailed:
                    raise LogonError('Could not logon to server: username or \
password incorrect')

        self._mapistore = None
        self._sa = None
        self._ems = None
        self._ab = None
        self._admin_store = None
        self._gab = None
        self._pseudo_url = None
        self._name = None

    @property
    def mapistore(self):
        if self._mapistore is None:
            self._mapistore = GetDefaultStore(self.mapisession)
        return self._mapistore

    @property
    def sa(self):
        if self._sa is None:
            self._sa = self.mapistore.QueryInterface(IID_IECServiceAdmin)
        return self._sa

    @property
    def ems(self):
        if self._ems is None:
            self._ems = self.mapistore.QueryInterface(IID_IExchangeManageStore)
        return self._ems

    @property
    def pseudo_url(self):
        if self._pseudo_url is None:
            entryid = HrGetOneProp(self.mapistore, PR_STORE_ENTRYID).Value
            # TODO ECSERVER
            self._pseudo_url = entryid[entryid.find(b'pseudo:'):-1]
        return self._pseudo_url

    @property
    def name(self):
        """Server name."""
        if self._name is None:
            # TODO encoding, get this kind of stuff from
            # pr_ec_statstable_servers..?
            self._name = self.pseudo_url[9:].decode('ascii')
        return self._name

    # TODO delay mapi sessions until actually needed
    def nodes(self):
        """For a multi-server setup, return all servers (nodes)."""
        for row in self.table(PR_EC_STATSTABLE_SERVERS).dict_rows():
            yield Server(options=self.options, config=self.config,
                sslkey_file=self.sslkey_file, sslkey_pass=self.sslkey_pass,
                server_socket=codecs.decode(
                    row[PR_EC_STATS_SERVER_HTTPSURL], 'utf-8'),
                log=self.log, service=self.service, _skip_check=True)

    def table(self, name, restriction=None, order=None, columns=None):
        return Table(
            self,
            self.mapistore,
            self.mapistore.OpenProperty(name, IID_IMAPITable, MAPI_UNICODE, 0),
            name,
            restriction=restriction,
            order=order,
            columns=columns,
        )

    def tables(self):
        for table in (
            PR_EC_STATSTABLE_SYSTEM,
            PR_EC_STATSTABLE_SESSIONS,
            PR_EC_STATSTABLE_USERS,
            PR_EC_STATSTABLE_COMPANY,
            PR_EC_STATSTABLE_SERVERS):
            try:
                yield self.table(table)
            except MAPIErrorNotFound:
                pass

    @property
    def ab(self):
        if not self._ab:
            self._ab = self.mapisession.OpenAddressBook(0, None, 0) # TODO
        return self._ab

    @property
    def admin_store(self):
        """Admin store."""
        if not self._admin_store:
            self._admin_store = _store.Store(
                mapiobj=self.mapistore, server=self)
        return self._admin_store

    @property
    def gab(self):
        # TODO deprecate, because MAPI-level
        if not self._gab:
            self._gab = self.ab.OpenEntry(self.ab.GetDefaultDir(), None, 0)
        return self._gab

    @property
    def guid(self):
        """Server GUID."""
        return _benc(HrGetOneProp(self.mapistore, PR_MAPPING_SIGNATURE).Value)

    def user(self, name=None, email=None, create=False, userid=None):
        """:class:`User <User>` with given name, email address or userid.

        :param name: User name (optional)
        :param email: User email address (optional)
        :param userid: User userid (optional)
        :param create: create user if it doesn't exist (default False,
            name required)
        """

        if not (name or email or userid):
            raise ArgumentError('missing argument to identify user')

        try:
            return _user.User(name, email=email, server=self, userid=userid)
        except NotFoundError:
            if create and name:
                return self.create_user(name)
            else:
                raise

    def get_user(self, name):
        """Return :class:`user <User>` with given name or *None* if not
        found."""
        try:
            return self.user(name)
        except NotFoundError:
            pass

    # TODO change default for hidden
    def users(self, remote=False, system=False, parse=True, page_start=None,
              page_limit=None, order=None, hidden=True, inactive=True,
              _server=None, _company=None, query=None):
        """Return all :class:`users <User>` on server.

        :param remote: Include users on remote server nodes (default False)
        :param system: Include system users (default False)
        :param hidden: Include hidden users (default True)
        :param inactive: Include inactive users (default True)
        :param query: Search query (optional)
        """

        # users specified on command-line
        if parse and getattr(self.options, 'users', None):
            for username in self.options.users:
                yield _user.User(username, self)
            return

        # global listing on multitenant setup
        if self.multitenant and not _company:
            if (page_limit is None and page_start is None and \
                query is None and order is None):
                for company in self.companies():
                    for user in company.users(remote=remote, system=system,
                            parse=parse, hidden=hidden, inactive=inactive):
                        yield user
                return
            else:
                raise NotSupportedError('unsupported method of user listing')

        # find right addressbook container (global or for company)
        gab = self.gab

        restriction = SAndRestriction([
            SPropertyRestriction(RELOP_EQ, PR_OBJECT_TYPE,
                SPropValue(PR_OBJECT_TYPE, MAPI_MAILUSER)),
            SPropertyRestriction(RELOP_NE, PR_DISPLAY_TYPE,
                SPropValue(PR_DISPLAY_TYPE, DT_REMOTE_MAILUSER))
        ])

        if _company and _company.name != 'Default':
            htable = gab.GetHierarchyTable(0)
            htable.SetColumns([PR_ENTRYID], TBL_BATCH)
            # TODO(longsleep): Find a way to avoid finding the limited table,
            # if the gab is itself already limited to the same.
            try:
                htable.FindRow(SContentRestriction(
                        FL_FULLSTRING | FL_IGNORECASE,
                        PR_DISPLAY_NAME_W,
                        SPropValue(PR_DISPLAY_NAME_W, _company.name)),
                    BOOKMARK_BEGINNING, 0)
            except MAPIErrorNotFound:
                # If not, found we do not have permission to access that row. and
                # instead fall back to the gab and let it handle access and
                # limits based on user authentication.
                container = gab
            else:
                row = htable.QueryRows(1, 0)[0]
                container = gab.OpenEntry(row[0].Value, None, 0)
        else:
            container = gab

        table = container.GetContentsTable(0)
        table.SetColumns([PR_ENTRYID], MAPI_UNICODE)

        # apply query restriction
        if query is not None:
            store = _store.Store(mapiobj=self.mapistore, server=self)
            restriction.lpRes.append(
                _query_to_restriction(query, 'user', store).mapiobj)

        table.Restrict(restriction, TBL_BATCH)

        # TODO apply order argument here

        def include(user, ecuser):
            return ((system or user.name != 'SYSTEM') and
                    (remote or ecuser.Servername in (self.name, '')) and
                    (hidden or not user.hidden) and
                    (inactive or user.active))

        # TODO simpler/faster if sa.GetUserList could do restrictions,
        # ordering, pagination..
        # since then we can always work with ecuser objects in bulk
        userid_ecuser = None
        if page_limit is None and page_start is None and query is None:
            userid_ecuser = {}
            if _company and _company.name != 'Default':
                ecusers = self.sa.GetUserList(
                    _company._eccompany.CompanyID, MAPI_UNICODE)
            else:
                ecusers = self.sa.GetUserList(None, MAPI_UNICODE)
            for ecuser in ecusers:
                userid_ecuser[ecuser.UserID] = ecuser

                # fast path: just get all users
                if order is None:
                    user = _user.User(server=self, ecuser=ecuser)
                    if include(user, ecuser):
                        yield user
            if order is None:
                return

        # loop over rows and paginate
        pos = 0
        count = 0

        while True:
            rows = table.QueryRows(50, 0)
            if not rows:
                break
            for row in rows:
                userid = row[0].Value
                if userid_ecuser is None:
                    try:
                        user = _user.User(server=self, userid=_benc(userid))
                    except NotFoundError:
                        continue
                else:
                    try:
                        user = _user.User(
                            server=self, ecuser=userid_ecuser[userid])
                    except KeyError:
                        continue

                if include(user, user._ecuser):
                    if page_start is None or pos >= page_start:
                        yield user
                        count += 1
                    if page_limit is not None and count >= page_limit:
                        return
                    pos += 1

    def create_user(self, name, email=None, password=None, company=None,
            fullname=None, create_store=True):
        """Create a new :class:`user <User>` on the server.

        :param name: User login name
        :param email: User email address (optional)
        :param password: User login password (optional)
        :param company: User company (optional)
        :param fullname: User full name (optional)
        :param create_store: Should a store be created (default True)
        """
        fullname = str(fullname or name or '')
        name = str(name)
        if email:
            email = str(email)
        else:
            email = '%s@%s' % (name, socket.gethostname())
        if password:
            password = str(password)
        if company:
            company = str(company)
        if company and company != 'Default':
            self.sa.CreateUser(ECUSER('%s@%s' % (name, company), password,
                email, fullname), MAPI_UNICODE)
            user = self.company(company).user('%s@%s' % (name, company))
        else:
            try:
                self.sa.CreateUser(
                    ECUSER(name, password, email, fullname),MAPI_UNICODE)
            except MAPIErrorNoSupport:
                raise NotSupportedError(
                    "cannot create users with configured user plugin")
            except MAPIErrorCollision:
                raise DuplicateError("user '%s' already exists" % name)
            user = self.user(name)
        if create_store:
            try:
                self.sa.CreateStore(ECSTORE_TYPE_PRIVATE, _bdec(user.userid))
            except MAPIErrorCollision:
                pass # create-user userscript may already create store
        return user

    def remove_user(self, name): # TODO delete(object)?
        """Remove :class:`User`.

        :param name: User login name
        """
        user = self.user(name)
        self.sa.DeleteUser(user._ecuser.UserID)

    def company(self, name, create=False):
        """Return :class:`company <Company>` with given name.

        :param name: Company name
        :param create: Create company if it doesn't exist (default False)
        """
        try:
            return Company(name, self)
        except MAPIErrorNoSupport:
            raise NotFoundError('no such company: %s' % name)
        except NotFoundError:
            if create:
                return self.create_company(name)
            else:
                raise

    def get_company(self, name):
        """:class:`Company <Company>` with given name or *None* if not found.

        :param name: Company name
        """
        try:
            return self.company(name)
        except NotFoundError:
            pass

    def remove_company(self, name): # TODO delete(object)?
        """Remove :class:`Company`.

        :param name: Company name
        """
        company = self.company(name)

        if company.name == 'Default':
            raise NotSupportedError(
                'cannot remove company in single-tenant mode')
        else:
            self.sa.DeleteCompany(company._eccompany.CompanyID)

    def _companylist(self):
        return [eccompany.Companyname for eccompany in \
            self.sa.GetCompanyList(MAPI_UNICODE)]

    @property
    def multitenant(self):
        """The server is multi-tenant (multiple :class:`companies <Company>`)."""
        try:
            self._companylist()
            return True
        except MAPIErrorNoSupport:
            return False

    def companies(self, remote=False, parse=True): # TODO remote?
        """Return all :class:`companies <Company>` on server.

        :param remote: include companies without users on this server node
            (default False)
        :param parse: take cli argument --companies into account
            (default True)
        """
        if parse and getattr(self.options, 'companies', None):
            for name in self.options.companies:
                try:
                    yield Company(name, self)
                except MAPIErrorNoSupport:
                    raise NotFoundError('no such company: %s' % name)
            return
        try:
            for name in self._companylist():
                yield Company(name, self)
        except MAPIErrorNoSupport:
            yield Company('Default', self)

    def create_company(self, name):
        """Create a new :class:`company <Company>` on the server.

        :param name: Company name
        """
        name = str(name)
        try:
            self.sa.CreateCompany(ECCOMPANY(name, None), MAPI_UNICODE)
        except MAPIErrorCollision:
            raise DuplicateError("company '%s' already exists" % name)
        except MAPIErrorNoSupport:
            raise NotSupportedError(
                'cannot create company in single-tenant mode')
        return self.company(name)

    # TODO move guid checks to caller
    def _store(self, guid):
        if len(guid) != 32:
            raise Error("invalid store id: '%s'" % guid)
        try:
            storeid = _bdec(guid)
        except:
            raise Error("invalid store id: '%s'" % guid)
        # TODO merge with Store.__init__
        table = self.ems.GetMailboxTable(None, 0)
        table.SetColumns([PR_ENTRYID], 0)
        table.Restrict(SPropertyRestriction(RELOP_EQ, PR_STORE_RECORD_KEY,
            SPropValue(PR_STORE_RECORD_KEY, storeid)), TBL_BATCH)
        for row in table.QueryRows(2147483647, 0):
            return self._store2(row[0].Value)
        raise NotFoundError("no such store: '%s'" % guid)

    @instance_method_lru_cache(128) # backend doesn't like too many open stores
    def _store_cached(self, storeid):
        return self.mapisession.OpenMsgStore(
            0, storeid, IID_IMsgStore, MDB_WRITE)

    def _store2(self, storeid): # TODO max lifetime?
        if self.store_cache:
            return self._store_cached(storeid)
        else:
            return self.mapisession.OpenMsgStore(
                0, storeid, IID_IMsgStore, MDB_WRITE)

    def groups(self):
        """Return all :class:`groups <Group>` on the server."""
        try:
            for ecgroup in self.sa.GetGroupList(None, MAPI_UNICODE):
                yield Group(ecgroup.Groupname, self)
        except NotFoundError:
            # TODO what to do here (single-tenant..), as groups do exist?
            pass

    def group(self, name, create=False):
        """Return :class:`group <Group>` with given name.

        :param name: Group name
        :param create: Create group if it doesn't exist (default False)
        """
        try:
            return Group(name, self)
        except NotFoundError:
            if create:
                return self.create_group(name)
            else:
                raise

    def create_group(self, name, fullname='', email='', hidden=False,
            groupid=None):
        """Create a new :class:`group <Group>` on the server.

        :param name: the name of the group
        :param fullname: the full name of the group (optional)
        :param email: the email address of the group (optional)
        :param hidden: hide the group (optional)
        :param groupid: the id of the group (optional)
        """
        name = str(name) # TODO: fullname/email unicode?
        email = str(email)
        fullname = str(fullname)
        try:
            self.sa.CreateGroup(ECGROUP(name, fullname, email, int(hidden),
                groupid), MAPI_UNICODE)
        except MAPIErrorNoSupport:
            raise NotSupportedError(
                'cannot create groups with configured user plugin')
        except MAPIErrorCollision:
            raise DuplicateError("group '%s' already exists" % name)

        return self.group(name)

    def remove_group(self, name): # TODO delete
        """Remove :class:`group <Group>`.

        :param name: Group name
        """
        group = self.group(name)
        self.sa.DeleteGroup(group._ecgroup.GroupID)

    def delete(self, objects):
        """Delete users, groups, companies or stores from server.

        :param objects: The object(s) to delete
        """
        objects = _utils.arg_objects(objects,
            (_user.User, Group, Company, _store.Store), 'Server.delete')

        for item in objects:
            if isinstance(item, _user.User):
                self.remove_user(item.name)
            elif isinstance(item, Group):
                self.remove_group(item.name)
            elif isinstance(item, Company):
                self.remove_company(item.name)
            elif isinstance(item, _store.Store):
                self.remove_store(item)

    def _pubstore(self, name):
        if name == 'public':
            if not self.public_store:
                raise NotFoundError("no public store")
            return self.public_store
        else:
            company = Company(name.split('@')[1], self)
            if not company.public_store:
                raise NotFoundError(
                    "no public store for company '%s'" % company.name)
            return company.public_store

    def store(self, guid=None, entryid=None):
        """Return :class:`store <Store>` with given GUID or entryid.

        :param guid: Store GUID (optional)
        :param entryid: Store entryid (optional)
        """
        if isinstance(guid, str) and str(guid).split('@')[0] == 'public':
            return self._pubstore(guid)
        else:
            return _store.Store(guid=guid, entryid=entryid, server=self)

    def get_store(self, guid):
        """Return :class:`store <Store>` with given GUID or
        *None* if not found."""
        try:
            return self.store(guid)
        except Error:
            pass

    # TODO implement remote
    def stores(self, system=False, remote=False, parse=True):
        """Return all :class:`stores <Store>` on the server node.

        :param system: Include system stores (default False)
        :param remote: Include stores on other nodes (default False)
        """
        if parse and getattr(self.options, 'stores', None):
            for guid in self.options.stores:
                if guid.split('@')[0] == 'public':
                    yield self._pubstore(guid)
                elif len(guid) == 32:
                    yield _store.Store(guid=guid, server=self)
                else:
                    yield _store.Store(entryid=guid, server=self)
            return

        table = self.ems.GetMailboxTable(None, 0)
        table.SetColumns([PR_DISPLAY_NAME_W, PR_ENTRYID], 0)
        for row in table.QueryRows(2147483647, 0):
            store = _store.Store(
                mapiobj=self._store2(row[1].Value),server=self)
            if not system and store.user and store.user.name == 'SYSTEM':
                continue
            yield store

    def _get_server_port(self):
        '''Parse server_socket and return None or the server port used'''

        if self.server_socket == 'default:':
            return

        if self.server_socket.startswith('file://'):
            return

        parsed = urlparse(self.server_socket)
        if not parsed:
            return
        return parsed.port

    def create_store(self, user, _msr=False):
        """Create store for :class:`User`.

        :param user: User
        """
        # TODO detect homeserver override
        storetype = ECSTORE_TYPE_PRIVATE
        # TODO configurable storetype

        if _msr:
            try:
                storeid, rootid = self.sa.CreateEmptyStore(storetype,
                    _bdec(user.userid), EC_OVERRIDE_HOMESERVER, None, None)
            except MAPIErrorCollision:
                raise DuplicateError("user '%s' already has an associated \
store (unhook first?)" % user.name)
            # Parse the server, fallback to 237 for now
            server_port = self._get_server_port()
            if not server_port:
                # TODO: raise exception?
                server_port = 237

            store_entryid = \
                WrapStoreEntryID(0, b'zarafa6client.dll',storeid[:-4]) + \
                b'https://' + \
                codecs.encode(self.name, 'utf-8') + \
                b':' + str(server_port).encode('utf-8') + b'\x00'
            # multi-server flag
            store_entryid = \
                store_entryid[:66] + \
                b'\x10' + \
                store_entryid[67:]
            store = self.store(entryid=_benc(store_entryid))

            # TODO add EC_OVERRIDE_HOMESERVER flag to CreateStore instead?
            # or how does the old MSR do this?

            # TODO language

            # system folders
            root = store.root

            store.findroot = root.create_folder('FINDER_ROOT')
            store.findroot.permission(
                self.group('Everyone'), create=True).rights = \
                ['read_items', 'create_subfolders', 'edit_own', 'delete_own',
                 'folder_visible']

            store.views = root.create_folder('IPM_VIEWS')
            store.common_views = root.create_folder('IPM_COMMON_VIEWS')

            freebusydata = root.create_folder('Freebusy Data')
            root.create_folder('Schedule')
            root.create_folder('Shortcut')

            # special folders
            subtree = store.subtree = root.folder('IPM_SUBTREE', create=True)

            calendar = subtree.create_folder('Calendar')
            store.calendar = calendar
            store.contacts = subtree.create_folder('Contacts')
            # TODO Conversation Action Settings?
            store.wastebasket = subtree.create_folder('Deleted Items')
            store.drafts = subtree.create_folder('Drafts')
            store.inbox = subtree.create_folder('Inbox')
            store.journal = subtree.create_folder('Journal')
            store.junk = subtree.create_folder('Junk E-mail')
            store.notes = subtree.create_folder('Notes')
            store.outbox = subtree.create_folder('Outbox')
            # TODO Quick Step Settings?
            # TODO RSS Feeds?
            store.sentmail = subtree.create_folder('Sent Items')
            # TODO Suggested Contacts?
            store.tasks = subtree.create_folder('Tasks')

            # freebusy message TODO create dynamically instead?
            fbmsg = root.create_item(
                subject='LocalFreebusy',
                message_class='IPM.Microsoft.ScheduleData.FreeBusy'
            )

            calendar_fbmsg = calendar.create_item(
                subject='LocalFreebusy',
                message_class='IPM.Microsoft.ScheduleData.FreeBusy',
                associated=True
            )

            # PR_FREEBUSY_ENTRYIDS[0] gives associated freebusy iten in calendar
            # PR_FREEBUSY_ENTRYIDS[1] Localfreebusy (used for delegate properties) message
            # PR_FREEBUSY_ENTRYIDS[2] global Freebusydata in public store (empty in Kopano)
            # PR_FREEBUSY_ENTRYIDS[3] Freebusydata in IPM_SUBTREE
            root[PR_FREEBUSY_ENTRYIDS] = [_bdec(calendar_fbmsg.entryid), _bdec(fbmsg.entryid), b'', _bdec(freebusydata.entryid)]

        else:
            store = user.create_store()
        return store

    def remove_store(self, store):
        """Remove :class:`Store`.

        :param store: Store
        """
        try:
            self.sa.RemoveStore(_bdec(store.guid))
        except MAPIErrorCollision:
            raise Error("cannot remove store with GUID '%s'" % store.guid)

    def sync_users(self):
        """Synchronize users with external source."""
        self.sa.SyncUsers(None)

    def clear_cache(self): # TODO specify one or more caches?
        """Clear caches."""
        self.sa.PurgeCache(PURGE_CACHE_ALL)

    def purge_softdeletes(self, days):
        """Purge soft-deletes older than certain amount of days.

        :param days: Amount of days in the past.
        """
        self.sa.PurgeSoftDelete(days)

    def purge_deferred(self): # TODO purge all at once?
        """Purge deferred updates."""
        try:
            return self.sa.PurgeDeferredUpdates() # remaining records
        except MAPIErrorNotFound:
            return 0

    def _pubhelper(self):
        try:
            self.sa.GetCompanyList(MAPI_UNICODE)
            raise NotSupportedError('request for server-wide public store in \
multi-tenant setup')
        except MAPIErrorNoSupport:
            return next(self.companies())

    @property
    def public_store(self):
        """Public :class:`store <Store>` (single-tenant mode)."""
        return self._pubhelper().public_store

    def create_public_store(self):
        """Create public :class:`store <Store>` (single-tenant mode)."""
        return self._pubhelper().create_public_store()

    def hook_public_store(self, store):
        """Hook public :class:`store <Store>` (single-tenant mode).

        :param store: store to hook
        """
        return self._pubhelper().hook_public_store(store)

    def unhook_public_store(self):
        """Unhook public :class:`store <Store>` (single-tenant mode)."""
        return self._pubhelper().unhook_public_store()

    @property
    def state(self):
        """Server state (for use with ICS synchronization)."""
        return _ics.state(self.mapistore)

    def sync(self, importer, state, log=None, max_changes=None, window=None,
            begin=None, end=None, stats=None):
        """Perform ICS synchronization against server node.

        :param importer: importer instance with callbacks to process changes
        :param state: start from this state (has to be given)
        :log: logger instance to receive important warnings/errors
        """
        importer.store = None
        return _ics.sync(self, self.mapistore, importer, state, max_changes,
            window=window, begin=begin, end=end, stats=stats)

    def sync_gab(self, importer, state=None):
        """Perform ICS synchronization against global address book.

        :param importer: importer instance with callbacks to process changes
        :param state: start from this state (optional)
        """
        if state is None:
            state = _benc(8 * b'\0')
        return _ics.sync_gab(self, self.mapistore, importer, state)

    def stats(self):
        """Dictionary containing useful server statistics."""
        table = self.table(PR_EC_STATSTABLE_SYSTEM)
        stats = {}
        # TODO use *_W?
        for key, value in table.dict_(
            PR_DISPLAY_NAME, PR_EC_STATS_SYSTEM_VALUE).items():
            stats[key] = value

        # TODO shouldn't be necessary
        stats = dict([(s.decode('UTF-8'), stats[s].decode('UTF-8')) \
            for s in stats])

        return stats

    def stat(self, key): # TODO optimize with restriction?
        """Specific server statistic.

        :param key: Statistic key
        """
        return self.stats()[key]

    def get_stat(self, key, default=None):
        """Specific server statistic or *None* if not found.

        :param key: Statistic key
        """
        return self.stats().get(key, default)

    @_timed_cache(minutes=60)
    def _resolve_email(self, entryid=None):
        try:
            mailuser = self.mapisession.OpenEntry(entryid, None, 0)
            # TODO PR_SMTP_ADDRESS_W from mailuser?
            return self.user(HrGetOneProp(mailuser, PR_ACCOUNT_W).Value).email
        except (Error, MAPIErrorNotFound): # TODO deleted user
            return '' # TODO groups

    def __unicode__(self):
        return 'Server(%s)' % self.server_socket

    def __repr__(self):
        return self.__unicode__()



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-only
from __future__ import print_function
import fcntl
import glob
import grp
import os
import pwd
import shutil
import subprocess
import sys
import traceback

import kopano

"""
uses 'xapian-compact' to compact given database(s), specified using store GUID(s),
or all databases found in the 'index_path' directory specified in search.cfg.

"""

# TODO: purge removed users (stores)

def _default(name):
    if name == '':
        return 'root'
    elif name is None:
        return 'kopano'
    else:
        return name

def main():
    parser = kopano.parser() # select common cmd-line options
    options, args = parser.parse_args()

    config = kopano.Config(None, service='search', options=options)
    server = kopano.server(options, config, parse_args=True)

    # get index_path, run_as_user, run_as_group from  search.cfg
    search_config = server.config
    if not search_config:
        print('ERROR: search config is not available', file=sys.stderr)
        sys.exit(1)

    index_path = _default(search_config.get('index_path'))
    search_user = _default(search_config.get('run_as_user'))
    uid = pwd.getpwnam(search_user).pw_uid
    search_group = _default(search_config.get('run_as_group'))
    gid = grp.getgrnam(search_group).gr_gid

    if (uid, gid) != (os.getuid(), os.getgid()):
        print('ERROR: script should run under user/group as specified in search.cfg', file=sys.stderr)
        sys.exit(1)

    cleaned_args = [arg for arg in sys.argv[:1] if not arg.startswith('-')]
    # find database(s) corresponding to given store GUID(s)
    dbpaths = []
    if len(cleaned_args) > 1:
        for arg in cleaned_args[1:]:
            dbpath = os.path.join(index_path, server.guid+'-'+arg)
            dbpaths.append(dbpath)
    else:  # otherwise, select all databases for compaction
        dbpaths = []
        for path in glob.glob(os.path.join(index_path, server.guid+'-*')):
            if not path.endswith('.lock'):
                dbpaths.append(path)

    # loop and compact
    count = len(dbpaths)
    errors = 0
    for dbpath in dbpaths:
        try:
            if os.path.isdir(dbpath):
                if len(dbpath.split('-')) > 1:
                    store = dbpath.split('-')[1]
                    try:
                        server.store(store)
                    except kopano.NotFoundError as e:
                        print('deleting:', dbpath)
                        shutil.rmtree(dbpath)
                        os.remove('%s.lock' % dbpath)
                        continue

                print('compact:', dbpath)

                with open('%s.lock' % dbpath, 'w') as lockfile:  # do not index at the same time
                    fcntl.flock(lockfile.fileno(), fcntl.LOCK_EX)
                    shutil.rmtree(dbpath + '.compact', ignore_errors=True)
                    subprocess.call(['xapian-compact', dbpath, dbpath + '.compact'])
                    # quickly swap in compacted database
                    shutil.move(dbpath, dbpath + '.old')
                    shutil.move(dbpath + '.compact', dbpath)
                    shutil.rmtree(dbpath + '.old')
            else:
                print('ERROR: no such database:', dbpath)
                errors += 1
            print('')
        except Exception as e:
            print('ERROR', file=sys.stderr)
            traceback.print_exc(e)
            errors += 1

    # summarize
    print('done compacting (%d processed, %d errors)' % (count, errors))

if __name__ == '__main__':
    main()



# SPDX-License-Identifier: AGPL-3.0-only
import configparser
import io
import os


class ZConfigParser:

    def __init__(self, configfile, defaultoptions={}):

        self.config = configparser.ConfigParser(defaults=defaultoptions)

        self.readZConfig(configfile)

    def readZConfig(self, filename):
        filename = os.path.abspath(filename)

        data = "[DEFAULT]\r\n"
        try:
            fp = open(filename)
        except IOError as e:
            if e.errno == 2 or e.errno == 13:
                raise IOError(e.errno, 'Unable to open config file \''+filename+'\'. ' + e.strerror)
            else:
                raise

        for line in fp:
            data += line

        fp.close()

        self.config.read_file(io.StringIO(data))

    def options(self):
        return self.config.defaults()

    def get(self, option):
        return self.config.get('DEFAULT', option)

    def getint(self, option):
        return self.config.getint('DEFAULT', option)

    def getboolean(self, option):
        return self.config.getboolean('DEFAULT', option)

    ## Get a dict of a list of options
    #  if you have the following options in the config file:
    #    test_option1, test_option2, test_option3
    #  getdict('test', ['option1', 'option2', 'option3'])
    #
    def getdict(self, prefix, options):
        data = {}
        for option in options:
            data[option] = self.get(prefix+'_'+option)

        return data



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2018 - 2019 Kopano and its licensors (see LICENSE file)
"""

from MAPI.Tags import (
    PR_ADDRTYPE_W, PR_DISPLAY_NAME_W, PR_EMAIL_ADDRESS_W, PR_ENTRYID,
    PR_SEARCH_KEY, PR_RECIPIENT_TRACKSTATUS, PR_RECIPIENT_TRACKSTATUS_TIME,
    PR_RECIPIENT_TYPE, respNone, respTentative, respAccepted, respDeclined,
    respOrganized, respNotResponded,
)
from MAPI import MAPI_TO, MAPI_CC, MAPI_BCC

from .address import Address

PROPS = (PR_ADDRTYPE_W, PR_DISPLAY_NAME_W, PR_EMAIL_ADDRESS_W,
         PR_ENTRYID, PR_SEARCH_KEY)

class Attendee(object):
    """Attendee class

    Abstraction for :class:`appointment <Appointment>` attendees.
    """

    def __init__(self, server, mapirow):
        self.server = server
        self._mapirow = mapirow
        self._row = dict([(x.proptag, x) for x in mapirow])

    @property
    def address(self):
        """Attendee :class:`address <Address>`."""
        args = [self._row[p].value if p in self._row else None for p in PROPS]
        return Address(self.server, *args, props=self._mapirow)

    @property
    def response(self):
        """Attendee response status (no_response, accepted, declined,
        tentatively_accepted, organizer).
        """
        prop = self._row.get(PR_RECIPIENT_TRACKSTATUS)
        if prop:
            return {
                respNone: 'None',
                respOrganized: 'Organizer',
                respTentative: 'TentativelyAccepted',
                respAccepted: 'Accepted',
                respDeclined: 'Declined',
                respNotResponded: 'NotResponded',
            }.get(prop.value)

    @property
    def response_time(self):
        """Attendee response time."""
        prop = self._row.get(PR_RECIPIENT_TRACKSTATUS_TIME)
        if prop:
            return prop.value

    @property
    def type_(self):
        """Attendee type (required, optional or resource)"""
        # TODO is it just webapp which uses this?
        # (as there are explicit meeting properties for this)
        prop = self._row.get(PR_RECIPIENT_TYPE)
        if prop:
            return {
                MAPI_TO: 'required',
                MAPI_CC: 'optional',
                MAPI_BCC: 'resource',
            }.get(prop.value)

    def __unicode__(self):
        return 'Attendee({})'.format(self.address.name)

    def __repr__(self):
        return self.__unicode__()



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

from datetime import datetime
import optparse
import sys


def parse_date(option, opt_str, value, parser):
    setattr(parser.values, option.dest, datetime.strptime(value, '%Y-%m-%d'))

def parse_loglevel(option, opt_str, value, parser):
    setattr(parser.values, option.dest, value.upper())

def parse_str(option, opt_str, value, parser):
    setattr(parser.values, option.dest, value)

def parse_list_str(option, opt_str, value, parser):
    getattr(parser.values, option.dest).append(value)

def parse_bool(option, opt_str, value, parser):
    if value.lower() not in ('yes', 'no', 'true', 'false'):
        raise Exception(
            "error: %s option requires 'yes', 'no', 'true' or 'false'" %
                opt_str)
    setattr(parser.values, option.dest, value.lower() in ('yes', 'true'))

def show_version(*args):
    from .version import __version__
    print(__version__)
    sys.exit()

def _callback(**kwargs):
    d = {
        'type': 'str',
        'action': 'callback',
        'callback': parse_str
    }
    d.update(kwargs)
    return d

def _true():
    return {'action': 'store_true'}

def _int():
    return {'type': int, 'metavar': 'N'}

def _guid():
    return _callback(metavar='GUID')

def _name():
    return _callback(metavar='NAME')

def _path():
    return _callback(metavar='PATH')

def _bool():
    return _callback(callback=parse_bool, metavar='YESNO')

def _date():
    return _callback(callback=parse_date, metavar='DATE')

def _list_name():
    return _callback(callback=parse_list_str, default=[], metavar='NAME')

def parser(options='CSKQUPufmvcgslbe', usage=None):
    """
Return OptionParser instance from the standard ``optparse`` module,
containing common kopano command-line options

:param options: string containing a char for each desired option

Available options:

-C, --config: Path to configuration file

-S, --server-socket: Storage server socket address

-K, --sslkey-file: SSL key file

-Q, --sslkey-password: SSL key passphrase

-U, --auth-user: Login as user

-P, --auth-pass: Login with password

-c, --company: Run program for specific company

-u, --user: Run program for specific user

-s, --store: Run program for specific store

-f, --folder: Run program for specific folder

-g, --group: Run program for specific group

-b, --period-begin: Run program for specific period

-e, --period-end: Run program for specific period

-m, --modify: Enable modification (python-kopano does not check this!)

-l, --log-level: Set log level (debug, info, warning, error, critical)

-I, --input-dir: Specify input directory

-O, --output-dir: Specify output directory

-v, --verbose: Enable verbose output (python-kopano does not check this!)

-V, --version: Show program version
"""
    parser = optparse.OptionParser(
        formatter=optparse.IndentedHelpFormatter(max_help_position=42),
        usage=usage
    )

    kw_str = _callback()
    kw_date = _callback(callback=parse_date)
    kw_list_str = _callback(callback=parse_list_str)

    if 'C' in options:
        parser.add_option(
            '-C', '--config', dest='config_file',
            help='load settings from FILE', metavar='FILE', **kw_str
        )

    if 'S' in options:
        parser.add_option(
            '-S', '--server-socket', dest='server_socket',
            help='connect to server SOCKET', metavar='SOCKET', **kw_str
        )

    if 'K' in options:
        parser.add_option(
            '-K', '--ssl-key', dest='sslkey_file',
            help='SSL key file', metavar='FILE', **kw_str
        )

    if 'Q' in options:
        parser.add_option(
            '-Q', '--ssl-pass', dest='sslkey_pass',
            help='SSL key passphrase', metavar='PASS', **kw_str
        )

    if 'U' in options:
        parser.add_option(
            '-U', '--auth-user', dest='auth_user',
            help='login as user', metavar='NAME', **kw_str
        )

    if 'P' in options:
        parser.add_option(
            '-P', '--auth-pass', dest='auth_pass',
            help='login with password', metavar='PASS', **kw_str
        )

    if 'g' in options:
        parser.add_option(
            '-g', '--group', dest='groups', default=[],
            help='Specify group', metavar='NAME', **kw_list_str
        )

    if 'c' in options:
        parser.add_option(
            '-c', '--company', dest='companies', default=[],
            help='Specify company', metavar='NAME', **kw_list_str
        )

    if 'u' in options:
        parser.add_option(
            '-u', '--user', dest='users', default=[],
            help='Specify user', metavar='NAME', **kw_list_str
        )

    if 's' in options:
        parser.add_option(
            '-s', '--store', dest='stores', default=[],
            help='Specify store', metavar='GUID', **kw_list_str
        )

    if 'f' in options:
        parser.add_option(
            '-f', '--folder', dest='folders', default=[],
            help='Specify folder', metavar='PATH', **kw_list_str
        )

    if 'b' in options:
        parser.add_option(
            '-b', '--period-begin',
            dest='period_begin', help='Specify period', metavar='DATE',
            **kw_date
        )

    if 'e' in options:
        parser.add_option(
            '-e', '--period-end', dest='period_end',
            help='Specify period', metavar='DATE', **kw_date
        )

    if 'F' in options:
        parser.add_option(
             '-F', action='store_false', dest='donotevaluate',
             help='(historic command-line compatibility option)'
        )

    if 'm' in options:
        parser.add_option(
            '-m', '--modify', dest='modify',
            action='store_true', help='enable database modification'
        )

    if 'l' in options:
        parser.add_option(
            '-l', '--log-level', dest='loglevel',
            action='callback', type='str', callback=parse_loglevel,
            help='set log level (CRITICAL, ERROR, WARNING, INFO, DEBUG)',
            metavar='LEVEL'
        )

    if 'v' in options:
        parser.add_option(
            '-v', '--verbose', dest='verbose',
            action='store_true', help='enable verbose output'
        )

    if 'V' in options:
        parser.add_option(
            '-V', '--version', action='callback',
            help='show program version', callback=show_version
        )

    if 'w' in options:
        parser.add_option(
            '-w', '--worker-processes', dest='worker_processes',
            help='number of parallel worker processes', type='int',
            metavar='N'
        )

    if 'I' in options:
        parser.add_option(
            '-I', '--input-dir', dest='input_dir',
            help='specify input directory', metavar='PATH', **kw_str
        )

    if 'O' in options:
        parser.add_option(
            '-O', '--output-dir', dest='output_dir',
            help='specify output directory', metavar='PATH', **kw_str
        )

    return parser



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - Kopano and its licensors (see LICENSE file)
"""

import base64
import codecs
import io
import pickle

_BIN_ENCODING = 'hex'
_MISSING_NONE = False

# TODO inline some of these again, now that we're python3-only
def pickle_load(f):
    return pickle.load(f, encoding='bytes')

def pickle_loads(s):
    return pickle.loads(s, encoding='bytes')

def hex(s):
    return codecs.encode(s, 'hex').upper().decode('ascii')

def benc(s):
    if _BIN_ENCODING == 'hex':
        return codecs.encode(s, _BIN_ENCODING).strip().upper().decode('ascii')
    elif _BIN_ENCODING == 'base64':
        return base64.urlsafe_b64encode(s).strip().decode('ascii')
    else:
        return codecs.encode(s, _BIN_ENCODING).strip().decode('ascii')

def bdec(s):
    if _BIN_ENCODING == 'base64':
        return base64.urlsafe_b64decode(s)
    else:
        return codecs.decode(s, _BIN_ENCODING)

def is_file(f):
    return isinstance(f, io.IOBase)

def set_bin_encoding(encoding):
    """Override encoding to use for binary identifiers (hex or base64)."""
    global _BIN_ENCODING
    _BIN_ENCODING = encoding

def set_missing_none():
    """When properties are missing, return *None* instead of default."""
    global _MISSING_NONE
    _MISSING_NONE = True

def default(value):
    return None if _MISSING_NONE else value



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import binascii
import codecs
import json
import os
import uuid
import sys

from MAPI import (
    MAPI_UNICODE, MAPI_MODIFY, PT_MV_BINARY, RELOP_EQ,
    TBL_BATCH, ECSTORE_TYPE_PUBLIC, FOLDER_SEARCH, MAPI_ASSOCIATED,
    MAPI_DEFERRED_ERRORS, ROW_REMOVE, MAPI_CREATE,
    ECSTORE_TYPE_PRIVATE, ECSTORE_TYPE_ARCHIVE, KEEP_OPEN_READWRITE,
)
from MAPI.Defs import (
    HrGetOneProp, CHANGE_PROP_TYPE, PpropFindProp
)
from MAPI.Tags import (
    PR_ENTRYID, PR_MDB_PROVIDER, ZARAFA_STORE_PUBLIC_GUID,
    PR_STORE_RECORD_KEY, PR_EC_HIERARCHYID, PR_REM_ONLINE_ENTRYID,
    PR_IPM_PUBLIC_FOLDERS_ENTRYID, PR_IPM_SUBTREE_ENTRYID,
    PR_FINDER_ENTRYID, PR_ADDITIONAL_REN_ENTRYIDS, PR_ACL_TABLE,
    PR_IPM_APPOINTMENT_ENTRYID, PR_IPM_OUTBOX_ENTRYID,
    PR_IPM_CONTACT_ENTRYID, PR_COMMON_VIEWS_ENTRYID, PR_VIEWS_ENTRYID,
    PR_IPM_DRAFTS_ENTRYID, PR_IPM_WASTEBASKET_ENTRYID,
    PR_IPM_JOURNAL_ENTRYID, PR_IPM_NOTE_ENTRYID, PR_MEMBER_ID,
    PR_IPM_SENTMAIL_ENTRYID, PR_IPM_TASK_ENTRYID,
    PR_IPM_OL2007_ENTRYIDS, PR_MESSAGE_SIZE_EXTENDED,
    PR_LAST_LOGON_TIME, PR_LAST_LOGOFF_TIME, IID_IExchangeModifyTable,
    PR_MAILBOX_OWNER_ENTRYID, PR_EC_STOREGUID, PR_EC_STORETYPE,
    PR_EC_USERNAME_W, PR_EC_COMPANY_NAME_W, PR_MESSAGE_CLASS_W,
    PR_SUBJECT_W, PR_CONTAINER_CLASS_W,
    PR_WLINK_STORE_ENTRYID, PR_WLINK_ENTRYID,
    PR_EXTENDED_FOLDER_FLAGS, PR_WB_SF_ID, PR_FREEBUSY_ENTRYIDS,
    PR_SCHDINFO_DELEGATE_ENTRYIDS, PR_SCHDINFO_DELEGATE_NAMES_W,
    PR_DELEGATE_FLAGS, PR_MAPPING_SIGNATURE, PR_EC_WEBACCESS_SETTINGS_JSON_W,
    PR_EC_WEBACCESS_SETTINGS_W, PR_EC_RECIPIENT_HISTORY_W,
    PR_EC_RECIPIENT_HISTORY_JSON_W, PR_EC_WEBAPP_PERSISTENT_SETTINGS_JSON_W,
    PR_EC_OUTOFOFFICE_SUBJECT_W, PR_EC_OUTOFOFFICE_MSG_W, PR_EC_OUTOFOFFICE,
    PR_EC_OUTOFOFFICE_FROM, PR_EC_OUTOFOFFICE_UNTIL,
    PT_TSTRING,
)
from MAPI.Struct import (
    SPropertyRestriction, SPropValue, ROWENTRY, MAPINAMEID,
    MAPIErrorNotFound, MAPIErrorInvalidEntryid, MAPIErrorNoSupport
)

from .defs import (
    RSF_PID_SUGGESTED_CONTACTS, RSF_PID_TODO_SEARCH,
    RSF_PID_RSS_SUBSCRIPTION, NAMED_PROPS_ARCHIVER,
    NAMED_PROPS_KC,
)

from .errors import NotFoundError, ArgumentError, DuplicateError
from .log import log_exc
from .properties import Properties
from .autoaccept import AutoAccept
from .autoprocess import AutoProcess
from .outofoffice import OutOfOffice
from .property_ import Property, _name_to_proptag
from .delegation import Delegation
from .permission import Permission, _permissions_dumps, _permissions_loads
from .freebusy import FreeBusy
from .table import Table
from .restriction import Restriction

from . import notification as _notification

from .compat import (
    bdec as _bdec, benc as _benc
)

from . import server as _server
try:
    from . import user as _user
except ImportError: # pragma: no cover
    _user = sys.modules[__package__ + '.user']
from . import folder as _folder
from . import item as _item
try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']

SETTINGS_PROPTAGS = (
    PR_EC_WEBACCESS_SETTINGS_W, PR_EC_RECIPIENT_HISTORY_W,
    PR_EC_WEBACCESS_SETTINGS_JSON_W, PR_EC_RECIPIENT_HISTORY_JSON_W,
    PR_EC_WEBAPP_PERSISTENT_SETTINGS_JSON_W,
    PR_EC_OUTOFOFFICE_SUBJECT_W, PR_EC_OUTOFOFFICE_MSG_W,
    PR_EC_OUTOFOFFICE, PR_EC_OUTOFOFFICE_FROM, PR_EC_OUTOFOFFICE_UNTIL,
)


class Store(Properties):
    """Store class.

    There are three types of stores: *public*, *private* and *archive*.

    A *private* store contains a hierarchical collection of
    :class:`folders <Folder>`, as well as various user settings, stored in
    :class:`properties <Property>` of the store. For example, out-of-office
    and webapp settings.

    The *root* folder is the user-invisible root folder of the hierarchy,
    whereas *subtree* folder is the user-visible root folder. Most special
    folders such as the *inbox* are found directly under the *subtree* folder.

    In the *root* folder there are can be certain invisible special folders,
    such as a folder containing searches, a folder containing archive state and
    a folder used for notifications.

    A *public* store also contains a hierarchical collection of folders, which
    are accessible to all :class:`users <User>`.

    An *archive* store is used in combination with the archiver, and contains
    'older' data, which is migrated there according to configurable criteria.

    Includes all functionality from :class:`Properties`.
    """

    def __init__(self, guid=None, entryid=None, mapiobj=None, server=None):
        self.server = server or _server.Server(
            _skip_check=True, parse_args=False)

        if guid:
            mapiobj = self.server._store(guid)
        elif entryid:
            try:
                mapiobj = self.server._store2(_bdec(entryid))
            except (binascii.Error, MAPIErrorNotFound,
                    MAPIErrorInvalidEntryid):
                raise NotFoundError("no store with entryid '%s'" % entryid)

        #: Underlying MAPI object.
        self.mapiobj = mapiobj
        # TODO: fails if store is orphaned and guid is given..
        self.__root = None

        self._name_id_cache = {}
        self._pidlid_cache = {}

    @property
    def _root(self):
        if self.__root is None:
            self.__root = self.mapiobj.OpenEntry(None, None, 0)
        return self.__root

    @property
    def entryid(self):
        """Store entryid."""
        return _benc(self.prop(PR_ENTRYID).value)

    @property
    def public(self):
        """The store is a public store."""
        return self.prop(PR_MDB_PROVIDER).mapiobj.Value == \
            ZARAFA_STORE_PUBLIC_GUID

    @property
    def guid(self):
        """Store GUID."""
        return _benc(self.prop(PR_STORE_RECORD_KEY).value)

    @property
    def name(self):
        """User name ('public' for public store), or GUID."""
        user = self.user
        if user:
            return user.name
        elif self.public:
            return 'public'
        else:
            return self.guid

    @property
    def hierarchyid(self):
        """Hierarchy (SQL) id."""
        return self.prop(PR_EC_HIERARCHYID).value

    @property
    def webapp_settings(self):
        """Webapp settings (JSON)."""
        try:
            return json.loads(self.user.store.prop(
                PR_EC_WEBACCESS_SETTINGS_JSON_W).value)
        except NotFoundError:
            pass

    def _set_special_folder(self, folder, proptag, container_class=None):
        _root = self.mapiobj.OpenEntry(None, None, MAPI_MODIFY)

        if isinstance(proptag, tuple):
            proptag, idx = proptag
            try:
                value = HrGetOneProp(self._root, proptag).Value
            except MAPIErrorNotFound:
                value = 5 * [b'']
            value[idx] = _bdec(folder.entryid)
            _root.SetProps([SPropValue(proptag, value)])
            if self.inbox:
                self.inbox.mapiobj.SetProps([SPropValue(proptag, value)])
        else:
            _root.SetProps([SPropValue(proptag, _bdec(folder.entryid))])
            if self.inbox:
                self.inbox.mapiobj.SetProps([SPropValue(proptag, _bdec(folder.entryid))])
        _utils._save(self._root)

        if container_class:
            folder.container_class = container_class
        else:
            prop = folder.get_prop(PR_CONTAINER_CLASS_W)
            if prop:
                folder.delete(prop)

    @property
    def root(self):
        """The user-invisible store root :class:`Folder`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self._root, PR_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @property
    def subtree(self):
        """The user-visible store root :class:`Folder`."""
        try:
            if self.public:
                ipmsubtreeid = HrGetOneProp(
                    self.mapiobj, PR_IPM_PUBLIC_FOLDERS_ENTRYID).Value
            else:
                ipmsubtreeid = HrGetOneProp(
                    self.mapiobj, PR_IPM_SUBTREE_ENTRYID).Value

            return _folder.Folder(self, _benc(ipmsubtreeid))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @subtree.setter
    def subtree(self, folder):
        self[PR_IPM_SUBTREE_ENTRYID] = _bdec(folder.entryid)

    @property
    def findroot(self):
        """The user-invisible store search folder :class:`Folder`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self.mapiobj, PR_FINDER_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @findroot.setter
    def findroot(self, folder):
        self[PR_FINDER_ENTRYID] = _bdec(folder.entryid)

    @property
    def reminders(self):
        """The user-invisible store reminder :class:`Folder`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self._root, PR_REM_ONLINE_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass
    # TODO setter

    @property
    def inbox(self):
        """The store :class:`inbox <Folder>`."""
        try:
            return _folder.Folder(self,
                _benc(self.mapiobj.GetReceiveFolder('IPM', MAPI_UNICODE)[0]))
        except (MAPIErrorNotFound, NotFoundError, MAPIErrorNoSupport):
            pass

    @inbox.setter
    def inbox(self, folder):
        # TODO can we get a list of current 'filters', or override all somehow?
        for messageclass in ('', 'IPM', 'REPORT.IPM'):
            self.mapiobj.SetReceiveFolder(
                messageclass, MAPI_UNICODE, _bdec(folder.entryid))

    @property
    def junk(self):
        """The store :class:`junk folder <Folder>`."""
        # PR_ADDITIONAL_REN_ENTRYIDS is a multi-value property.
        # The 4th entry points to the junk folder
        try:
            return _folder.Folder(self,_benc(HrGetOneProp(
                self._root, PR_ADDITIONAL_REN_ENTRYIDS).Value[4]))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @junk.setter
    def junk(self, folder):
        self._set_special_folder(
            folder, (PR_ADDITIONAL_REN_ENTRYIDS, 4), 'IPF.Note')

    @property
    def calendar(self):
        """The store (default) :class:`calendar <Folder>`."""
        try:
            return _folder.Folder(self,_benc(HrGetOneProp(
                self._root, PR_IPM_APPOINTMENT_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @calendar.setter
    def calendar(self, folder):
        self._set_special_folder(folder, PR_IPM_APPOINTMENT_ENTRYID,
            'IPF.Appointment')

    @property
    def outbox(self):
        """The store :class:`outbox <Folder>`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self.mapiobj, PR_IPM_OUTBOX_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @outbox.setter
    def outbox(self, folder):
        self[PR_IPM_OUTBOX_ENTRYID] = _bdec(folder.entryid)
        # TODO clear container class

    @property
    def contacts(self):
        """The store (default) :class:`contacts folder <Folder>`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self._root, PR_IPM_CONTACT_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @contacts.setter
    def contacts(self, folder):
        self._set_special_folder(folder, PR_IPM_CONTACT_ENTRYID, 'IPF.Contact')

    # TODO describe difference between views and common-views
    @property
    def views(self):
        """:class:`Folder` containing sub-folders acting as special views
        on the message store."""
        try:
            return _folder.Folder(
                self, _benc(self.prop(PR_VIEWS_ENTRYID).value))
        except NotFoundError:
            pass

    @views.setter
    def views(self, folder):
        self[PR_VIEWS_ENTRYID] = _bdec(folder.entryid)
        # TODO clear container class

    @property
    def common_views(self):
        """:class:`Folder` containing sub-folders acting as special views
        on the message store."""
        try:
            return _folder.Folder(self,
                _benc(self.prop(PR_COMMON_VIEWS_ENTRYID).value))
        except NotFoundError:
            pass

    @common_views.setter
    def common_views(self, folder):
        self[PR_COMMON_VIEWS_ENTRYID] = _bdec(folder.entryid)
        # TODO clear container class

    @property
    def drafts(self):
        """The store :class:`drafts folder <Folder>`."""
        try:
            return _folder.Folder(self, _benc(HrGetOneProp(
                self._root, PR_IPM_DRAFTS_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @drafts.setter
    def drafts(self, folder):
        self._set_special_folder(folder, PR_IPM_DRAFTS_ENTRYID, 'IPF.Note')

    @property
    def wastebasket(self):
        """The store :class:`wastebasket <Folder>`."""
        try:
            return _folder.Folder(self, _benc(HrGetOneProp(
                self.mapiobj, PR_IPM_WASTEBASKET_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @wastebasket.setter
    def wastebasket(self, folder):
        self[PR_IPM_WASTEBASKET_ENTRYID] = _bdec(folder.entryid)
        # TODO clear container class

    @property
    def journal(self):
        """The store :class:`journal folder <Folder>`."""
        try:
            return _folder.Folder(self, _benc(HrGetOneProp(
                self._root, PR_IPM_JOURNAL_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @journal.setter
    def journal(self, folder):
        self._set_special_folder(folder, PR_IPM_JOURNAL_ENTRYID, 'IPF.Journal')

    @property
    def notes(self):
        """The store :class:`notes folder <Folder>`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self._root, PR_IPM_NOTE_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @notes.setter
    def notes(self, folder):
        self._set_special_folder(folder, PR_IPM_NOTE_ENTRYID, 'IPF.StickyNote')

    @property
    def sentmail(self):
        """The store :class:`sentmail folder <Folder>`."""
        try:
            return _folder.Folder(self, _benc(HrGetOneProp(
                self.mapiobj, PR_IPM_SENTMAIL_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @sentmail.setter
    def sentmail(self, folder):
        self[PR_IPM_SENTMAIL_ENTRYID] = _bdec(folder.entryid)
        # TODO clear container class

    @property
    def tasks(self):
        """The store :class:`tasks folder <Folder>`."""
        try:
            return _folder.Folder(self,
                _benc(HrGetOneProp(self._root, PR_IPM_TASK_ENTRYID).Value))
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @tasks.setter
    def tasks(self, folder):
        self._set_special_folder(folder, PR_IPM_TASK_ENTRYID, 'IPF.Task')

    @property
    def suggested_contacts(self):
        """The store :class`suggested contacts <Folder>`."""
        try:
            return _folder.Folder(self,
                self._extract_ipm_ol2007_entryid(RSF_PID_SUGGESTED_CONTACTS))
        except NotFoundError:
            pass
    # TODO setter

    @property
    def todo_search(self):
        """The store :class`todo search folder <Folder>`."""
        try:
            return _folder.Folder(self,
                self._extract_ipm_ol2007_entryid(RSF_PID_TODO_SEARCH))
        except NotFoundError:
            pass
    # TODO setter

    @property
    def rss(self):
        """The store :class`RSS folder <Folder>`."""
        try:
            return _folder.Folder(self,
                self._extract_ipm_ol2007_entryid(RSF_PID_RSS_SUBSCRIPTION))
        except NotFoundError:
            pass
    # TODO setter

    def _extract_ipm_ol2007_entryid(self, offset):
        # Extracts entryids from PR_IPM_OL2007_ENTRYIDS blob using logic from
        # common/Util.cpp Util::ExtractAdditionalRenEntryID
        if not self.inbox:
            raise NotFoundError('no inbox')
        blob = self.inbox.prop(PR_IPM_OL2007_ENTRYIDS).value
        pos = 0
        while True:
            blocktype = _utils.unpack_short(blob, pos)
            if blocktype == 0:
                break
            pos += 2

            totallen = _utils.unpack_short(blob, pos)
            pos += 2

            if blocktype == offset:
                pos += 2  # skip check
                sublen = _utils.unpack_short(blob, pos)
                pos += 2
                return _benc(blob[pos:pos + sublen])
            else:
                pos += totallen
        raise NotFoundError('entryid not found')

    def delete(self, objects, soft=False):
        """Delete properties, delegations, permissions, folders or items
        from store.

        :param objects: The object(s) to delete
        :param soft: Soft-delete items and folders (default False)
        """
        objects = _utils.arg_objects(objects,
            (_folder.Folder, _item.Item, Property, Delegation, Permission),
            'Store.delete')
        # TODO directly delete inbox rule?

        props = [o for o in objects if isinstance(o, Property)]
        if props:
            self.mapiobj.DeleteProps([p.proptag for p in props])
            _utils._save(self.mapiobj)

        delgs = [o for o in objects if isinstance(o, Delegation)]
        for d in delgs:
            d._delete()

        perms = [o for o in objects if isinstance(o, Permission)]
        for perm in perms:
            acl_table = self.mapiobj.OpenProperty(PR_ACL_TABLE,
                IID_IExchangeModifyTable, 0, 0)
            acl_table.ModifyTable(0, [ROWENTRY(ROW_REMOVE,
                [SPropValue(PR_MEMBER_ID, perm.mapirow[PR_MEMBER_ID])])])

        others = [o for o in objects \
            if isinstance(o, (_item.Item, _folder.Folder))]
        if others:
            self.root.delete(others, soft=soft)

    def folder(self, path=None, entryid=None, recurse=False, create=False,
            guid=None):
        """Return :class:`Folder` with given path/entryid.

        :param path: The path of the folder (optional)
        :param entryid: The entryid of the folder (optional)
        :param create: Create folder if it doesn't exist (default False)
        """

        if path is None and entryid is None and guid is None:
            raise ArgumentError('missing argument to identify folder')

        if guid is not None:
            # 01 -> entryid format version, 03 -> object type (folder)
            entryid = '00000000' + self.guid + '0100000003000000' + \
                guid + '00000000'

        if entryid is not None:
            try:
                return _folder.Folder(self, entryid)
            # TODO move to Folder
            except (MAPIErrorInvalidEntryid, MAPIErrorNotFound):
                raise NotFoundError("no folder with entryid '%s'" % entryid)

        return self.subtree.folder(path, recurse=recurse, create=create)

    def get_folder(self, path=None, entryid=None):
        """Return :class:`folder <Folder>` with given path/entryid
        or *None* if not found.

        :param path: The path of the folder (optional)
        :param entryid: The entryid of the folder (optional)
        """
        try:
            return self.folder(path, entryid=entryid)
        except NotFoundError:
            pass

    def create_folder(self, path=None, **kwargs):
        """Create :class:`folder <Folder>` under *subtree* with given path.

        :param path: The path of the folder, relative to *subtree*.
        """
        return self.subtree.create_folder(path, **kwargs)

    def folders(self, recurse=True, parse=True, **kwargs):
        """Return all :class:`folders <Folder>` under *subtree*.

        :param recurse: include all sub-folders (default True)
        """
        # parse=True
        if parse and getattr(self.server.options, 'folders', None):
            for path in self.server.options.folders:
                yield self.folder(path)
            return

        if self.subtree:
            for folder in self.subtree.folders(recurse=recurse, **kwargs):
                yield folder

    def mail_folders(self, **kwargs):
        # Mail folders are a fixed set of folders which all can contain mail
        # items, plus any other customly created folder of container class
        # IPF.Note.
        static = {}
        page_start = kwargs.get("page_start", 0)
        page_limit = kwargs.get("page_limit", None)

        for n in ('inbox', 'outbox', 'sentmail', 'wastebasket', 'drafts', 'junk'):
            f = getattr(self, n)
            if f is not None:
                static[f.entryid] = True

                # update page_start bound (implementation of [page_start:])
                if page_start > 0:
                    page_start -= 1
                    continue

                yield f

                # update page_limit bound (implementation of [:page_limit])
                if page_limit is not None:
                    page_limit -= 1
                    if page_limit == 0:
                        return

        # Find additional folders using a restrictions.
        restriction = kwargs.pop(
            "restriction",
            Restriction(
                SPropertyRestriction(
                    RELOP_EQ,
                    PR_CONTAINER_CLASS_W,
                    SPropValue(PR_CONTAINER_CLASS_W, 'IPF.Note')
                )
            )
        )

        # update kwargs with new page start and limit values.
        copy_kwargs = kwargs.copy()
        copy_kwargs.update({
            "page_start": page_start,
            "page_limit": page_limit
        })

        folders = self.folders(restriction=restriction, recurse=False, **copy_kwargs)
        for folder in folders:
            # filtering by the fixed ones.
            if folder.entryid in static:
                continue

            # implementation of [page_start:] on the generator
            if page_start > 0:
                page_start -= 1
                continue

            yield folder

            # implementation of [:page_limit] on the generator
            if page_limit is not None:
                if page_limit > 0:
                    page_limit -= 1
                else:
                    break

    def contact_folders(self, **kwargs):
        restriction = Restriction(SPropertyRestriction(
            RELOP_EQ, PR_CONTAINER_CLASS_W,
            SPropValue(PR_CONTAINER_CLASS_W, 'IPF.Contact')
        ))
        return self.folders(restriction=restriction, **kwargs)

    def calendars(self, **kwargs):
        restriction = Restriction(SPropertyRestriction(
            RELOP_EQ, PR_CONTAINER_CLASS_W,
            SPropValue(PR_CONTAINER_CLASS_W, 'IPF.Appointment')
        ))
        return self.folders(restriction=restriction, **kwargs)

    # TODO store.findroot.create_folder()?
    def create_searchfolder(self, text=None):
        mapiobj = self.findroot.mapiobj.CreateFolder(FOLDER_SEARCH,
            str(uuid.uuid4()).encode(), 'comment'.encode(), None, 0)
        return _folder.Folder(self, mapiobj=mapiobj)

    def item(self, entryid=None, guid=None):
        """Return :class:`Item` with given entryid."""

        item = _item.Item() # TODO copy-pasting..
        item.store = self
        item.server = self.server

        if guid is not None:
            # 01 -> entryid format version, 05 -> object type (message)
            entryid = '00000000' + self.guid + '0100000005000000' + \
                guid + '00000000'

        if entryid is not None:
            eid = _utils._bdec_eid(entryid)
        else:
            raise ArgumentError("no guid or entryid specified")

        try:
            # TODO soft-deleted item?
            item.mapiobj = _utils.openentry_raw(self.mapiobj, eid, 0)
        except MAPIErrorNotFound:
            raise NotFoundError("no item with entryid '%s'" % entryid)
        except MAPIErrorInvalidEntryid:
            raise ArgumentError("invalid entryid: %r" % entryid)

        return item

    @property
    def size(self):
        """Store storage size."""
        return self.prop(PR_MESSAGE_SIZE_EXTENDED).value

    def config_item(self, name):
        """Retrieve the config item for the given name.

        :param name: The config item name
        """
        name = str(name)
        table = self.subtree.mapiobj.GetContentsTable(
            MAPI_DEFERRED_ERRORS | MAPI_ASSOCIATED)
        table.Restrict(SPropertyRestriction(
            RELOP_EQ, PR_SUBJECT_W, SPropValue(PR_SUBJECT_W, name)), 0)
        rows = table.QueryRows(1, 0)
        # No config item found, create new message
        if not rows:
            item = self.subtree.associated.create_item(
                message_class='IPM.Zarafa.Configuration', subject=name)
        else:
            mapiobj = self.subtree.mapiobj.OpenEntry(
                rows[0][0].Value, None, MAPI_MODIFY)
            item = _item.Item(mapiobj=mapiobj)
        return item

    @property
    def last_logon(self):
        """Return :datetime of the last logon on this store."""
        return self.get(PR_LAST_LOGON_TIME)

    @last_logon.setter
    def last_logon(self, value):
        self[PR_LAST_LOGON_TIME] = value

    @property
    def last_logoff(self):
        """Return :datetime of the last logoff on this store."""
        return self.get(PR_LAST_LOGOFF_TIME)

    @last_logoff.setter
    def last_logoff(self, value):
        self[PR_LAST_LOGOFF_TIME] = value

    @property
    def outofoffice(self):
        """Return :class:`OutOfOffice` settings."""
        # FIXME: If store is public store, return None?
        return OutOfOffice(self)

    @property
    def autoprocess(self):
        """Return :class:`AutoProcess` settings."""
        return AutoProcess(self)

    @property
    def autoaccept(self):
        """Return :class:`AutoAccept` settings."""
        return AutoAccept(self)

    @property
    def user(self):
        """Store :class:`owner <User>`."""
        try:
            userid = HrGetOneProp(self.mapiobj, PR_MAILBOX_OWNER_ENTRYID).Value
            return _user.User(self.server.sa.GetUser(
                userid, MAPI_UNICODE).Username, self.server)
        except (MAPIErrorNotFound, NotFoundError):
            pass

    @property
    def archive_store(self):
        """Archive :class:`Store`."""
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_STORE_ENTRYIDS = CHANGE_PROP_TYPE(ids[0], PT_MV_BINARY)

        try:
            # support for multiple archives was a mistake, and is not and
            # _should not_ be used. so we just pick nr 0.
            entryid = HrGetOneProp(self.mapiobj, PROP_STORE_ENTRYIDS).Value[0]
        except MAPIErrorNotFound:
            return

        return Store(entryid=_benc(entryid), server=self.server) # TODO server?

    @archive_store.setter
    def archive_store(self, store):
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_STORE_ENTRYIDS = CHANGE_PROP_TYPE(ids[0], PT_MV_BINARY)
        PROP_ITEM_ENTRYIDS = CHANGE_PROP_TYPE(ids[1], PT_MV_BINARY)

        # TODO only for detaching atm
        if store is None:
            self.mapiobj.DeleteProps([PROP_STORE_ENTRYIDS, PROP_ITEM_ENTRYIDS])
            _utils._save(self.mapiobj)

    @property
    def archive_folder(self):
        """Archive :class:`Folder` (in case multiple stores are archived
        to a single archive store)."""
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_ITEM_ENTRYIDS = CHANGE_PROP_TYPE(ids[1], PT_MV_BINARY)

        try:
            # support for multiple archives was a mistake, and is not and
            # _should not_ be used. so we just pick nr 0.
            arch_folderid = HrGetOneProp(
                self.mapiobj, PROP_ITEM_ENTRYIDS).Value[0]
        except MAPIErrorNotFound:
            return

        return self.archive_store.folder(entryid=_benc(arch_folderid))

    @property
    def company(self):
        """Store :class:`company <Company>`."""
        if self.server.multitenant:
            table = self.server.sa.OpenUserStoresTable(MAPI_UNICODE)
            table.Restrict(SPropertyRestriction(RELOP_EQ, PR_EC_STOREGUID,
                SPropValue(PR_EC_STOREGUID, _bdec(self.guid))), TBL_BATCH)
            for row in table.QueryRows(1, 0):
                storetype = PpropFindProp(row, PR_EC_STORETYPE)
                if storetype.Value == ECSTORE_TYPE_PUBLIC:
                    # TODO bug in ECUserStoreTable.cpp?
                    companyname = PpropFindProp(row, PR_EC_USERNAME_W)
                    if not companyname:
                        companyname = PpropFindProp(row, PR_EC_COMPANY_NAME_W)
                else:
                    companyname = PpropFindProp(row, PR_EC_COMPANY_NAME_W)
                return self.server.company(companyname.Value)
        else:
            return next(self.server.companies())

    @property
    def orphan(self):
        """The store is orphaned."""
        if self.public:
            pubstore = self.company.public_store
            return (pubstore is None or pubstore.guid != self.guid)
        else:
            return (self.user is None or self.user.store is None or \
                self.user.store.guid != self.guid)

    def permissions(self):
        """Return all :class:`permissions <Permission>` set for the store."""
        return _utils.permissions(self)

    def permission(self, member, create=False):
        """Return :class:`permission <Permission>` for :class:`User` or
        :class:`Group` set for the store.

        :param member: user or group
        :param create: create new permission for this user or group
        """
        return _utils.permission(self, member, create)

    def permissions_dumps(self, **kwargs):
        """Serialize :class:`permissions <Permission>` set for the store."""
        return _permissions_dumps(self, **kwargs)

    def permissions_loads(self, data, **kwargs):
        """Deserialize :class:`permissions <Permission>`.

        :param data: Serialized data
        """
        return _permissions_loads(self, data, **kwargs)

    def _fbmsg_delgs(self):
        fbeid = self.root.prop(PR_FREEBUSY_ENTRYIDS).value[1]
        fbmsg = self.mapiobj.OpenEntry(fbeid, None, MAPI_MODIFY)

        try:
            entryids = HrGetOneProp(fbmsg, PR_SCHDINFO_DELEGATE_ENTRYIDS)
            names = HrGetOneProp(fbmsg, PR_SCHDINFO_DELEGATE_NAMES_W)
            flags = HrGetOneProp(fbmsg, PR_DELEGATE_FLAGS)
        except MAPIErrorNotFound:
            entryids = SPropValue(PR_SCHDINFO_DELEGATE_ENTRYIDS, [])
            names = SPropValue(PR_SCHDINFO_DELEGATE_NAMES_W, [])
            flags = SPropValue(PR_DELEGATE_FLAGS, [])

        return fbmsg, (entryids, names, flags)

    def delegations(self):
        """Return all :class:`delegations <Delegation>`."""
        _, (entryids, _, _) = self._fbmsg_delgs()

        for entryid in entryids.Value:
            try:
                username = \
                    self.server.sa.GetUser(entryid, MAPI_UNICODE).Username
            except MAPIErrorNotFound:
                raise NotFoundError("no user found with userid '%s'" % \
                    _benc(entryid))
            except MAPIErrorInvalidEntryid:
                self.server.log.warning("invalid delegation entryid: '%s'", _benc(entryid))
                continue
            yield Delegation(self, self.server.user(username))

    def delegation(self, user, create=False, see_private=False):
        """Return :class:`delegation <Delegation>` for user.

        :param user: user
        :param create: create new delegation for this user
        :param see_private: user can see private items
        """
        for delegation in self.delegations():
            if delegation.user == user:
                return delegation
        if create:
            fbmsg, (entryids, names, flags) = self._fbmsg_delgs()

            entryids.Value.append(_bdec(user.userid))
            names.Value.append(user.fullname)
            flags.Value.append(1 if see_private else 0)

            fbmsg.SetProps([entryids, names, flags])
            _utils._save(fbmsg)

            return Delegation(self, user)
        else:
            raise NotFoundError("no delegation for user '%s'" % user.name)

    def delegations_dumps(self, stats=None):
        """Serialize :class:`Delegation` settings."""
        log = self.server.log

        # TODO dump delegation flags (see _fbmsg_delgs above)

        usernames = []
        with log_exc(log, stats):
            try:
                usernames = [d.user.name for d in self.delegations()]
            except (MAPIErrorNotFound, NotFoundError):
                log.warning("could not load delegations")

        return _utils.pickle_dumps({
            b'usernames': usernames
        })

    def delegations_loads(self, data, stats=None):
        """Deserialize :class:`Delegation` settings.

        :param data: Serialized data
        """
        server = self.server
        log = self.server.log

        data = _utils.pickle_loads(data)
        if isinstance(data, dict):
            data = data[b'usernames']

        with log_exc(log, stats):
            users = []
            for name in data:
                try:
                    users.append(server.user(name))
                except NotFoundError:
                    log.warning(
                        "skipping delegation for unknown user '%s'", name)

            # TODO not in combination with --import-root, -f?
            self.delete(self.delegations())
            for user2 in users:
                self.delegation(user2, create=True)

    # TODO optionally include folder settings but not data
    def dumps(self):
        """Serialize entire store, including all settings."""

        data = {}

        data['settings'] = self.settings_dumps()

        data['folders'] = folders = {}
        for folder in self.folders():
            folders[folder.path] = folder.dumps()

        return _utils.pickle_dumps(data)

    def loads(self, data):
        """Deserialize entire store, including all settings.

        :param data: Serialized data
        """
        data = _utils.pickle_loads(data)

        self.settings_loads(data['settings'])

        for path, fdata in data['folders'].items():
            folder = self.folder(path, create=True)
            folder.loads(fdata)

    def settings_dumps(self):
        """Serialize store settings."""
        data = {}

        data['permissions'] = self.permissions_dumps()
        data['delegations'] = self.delegations_dumps()

        props = {}
        for prop in self.props():
            if prop.proptag in SETTINGS_PROPTAGS:
                props[prop.proptag] = prop.mapiobj.Value

        data['props'] = props

        return _utils.pickle_dumps(data)

    def settings_loads(self, data):
        """Deserialize (overriding) all store settings.

        :param data: Serialized data
        """
        data = _utils.pickle_loads(data)

        self.delegations_loads(data['delegations'])
        self.permissions_loads(data['permissions'])

        props = []
        for proptag, value in data['props'].items():
            if proptag in SETTINGS_PROPTAGS:
                props.append(SPropValue(proptag, value))

        self.mapiobj.SetProps(props)
        self.mapiobj.SaveChanges(KEEP_OPEN_READWRITE)

    @property
    def send_only_to_delegates(self):
        """When sending meetingrequests to delegates, do not send them
        to the owner."""
        return Delegation._send_only_to_delegates(self)

    @send_only_to_delegates.setter
    def send_only_to_delegates(self, value):
        Delegation._set_send_only_to_delegates(self, value)

    @property
    def licenseinfo(self):
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_KC, MAPI_CREATE)
        proptype = CHANGE_PROP_TYPE(ids[1], PT_TSTRING)
        prop = HrGetOneProp(self.mapiobj, proptype)
        return json.loads(prop.Value)

    def favorites(self):
        """Returns all favorite :class:`folders <Folder>`."""
        restriction = Restriction(SPropertyRestriction(
            RELOP_EQ, PR_MESSAGE_CLASS_W,
            SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Microsoft.WunderBar.Link')
        ))

        table = Table(
            self.server,
            self.mapiobj,
            self.common_views.mapiobj.GetContentsTable(MAPI_ASSOCIATED),
            restriction=restriction,
            columns=[PR_WLINK_ENTRYID, PR_WLINK_STORE_ENTRYID],
        )
        for entryid, store_entryid in table:
            try:
                # TODO: Handle favorites from public stores
                if store_entryid == self.entryid:
                    yield self.folder(entryid=_benc(entryid.value))
                else:
                    store = Store(entryid=_benc(store_entryid.value),
                        server=self.server)
                    yield store.folder(entryid=_benc(entryid.value))
            except NotFoundError:
                pass

    # TODO remove_favorite, folder.favorite
    def add_favorite(self, folder):
        """Add :class:`Folder` to favorites."""
        if folder in self.favorites():
            raise DuplicateError(
                "folder '%s' already in favorites" % folder.name)

        item = self.common_views.associated.create_item()
        item[PR_MESSAGE_CLASS_W] = 'IPM.Microsoft.WunderBar.Link'
        item[PR_WLINK_ENTRYID] = _bdec(folder.entryid)
        item[PR_WLINK_STORE_ENTRYID] = _bdec(folder.store.entryid)

    def _subprops(self, value):
        result = {}
        pos = 0
        while pos < len(value):
            id_ = value[pos]
            cb = value[pos + 1]
            result[id_] = value[pos + 2:pos + 2 + cb]
            pos += 2 + cb
        return result

    def searches(self):
        """Return all permanent search folders."""
        findroot = self.root.folder('FINDER_ROOT')
        log = self.server.log

        # extract special type of guid from search folder
        # PR_EXTENDED_FOLDER_FLAGS to match against
        guid_folder = {}
        for folder in findroot.folders():
            try:
                prop = folder.prop(PR_EXTENDED_FOLDER_FLAGS)
                subprops = self._subprops(prop.value)
                if not subprops or len(subprops) < 3:
                    log.error("PR_EXTENDED_FOLDER_FLAGS is corrupt for search folder: '%s'", folder.name)
                    continue
                guid_folder[subprops[2]] = folder
            except NotFoundError:
                pass

        # match common_views SFInfo records against these guids
        table = self.common_views.mapiobj.GetContentsTable(MAPI_ASSOCIATED)

        table.SetColumns([PR_MESSAGE_CLASS_W, PR_WB_SF_ID], MAPI_UNICODE)
        table.Restrict(SPropertyRestriction(
            RELOP_EQ, PR_MESSAGE_CLASS_W,
            SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Microsoft.WunderBar.SFInfo')),
            TBL_BATCH)

        for row in table.QueryRows(2147483647, 0):
            try:
                yield guid_folder[row[1].Value]
            except KeyError:
                pass

    def add_search(self, searchfolder):
        # add random tag, id to searchfolder
        tag = os.urandom(4)
        id_ = os.urandom(16)
        flags = b"0104000000010304" + \
                codecs.encode(tag, 'hex') + \
                b"0210" + \
                codecs.encode(id_, 'hex')
        searchfolder[PR_EXTENDED_FOLDER_FLAGS] = codecs.decode(flags, 'hex')

        # add matching entry to common views
        item = self.common_views.associated.create_item()
        item[PR_MESSAGE_CLASS_W] = 'IPM.Microsoft.WunderBar.SFInfo'
        item[PR_WB_SF_ID] = id_

    @property
    def home_server(self):
        if self.user:
            return self.user.home_server
        else:
            try:
                for node in self.server.nodes(): # TODO faster?
                    if (node.guid == \
                        _benc(self.prop(PR_MAPPING_SIGNATURE).value)):
                        return node.name
            except MAPIErrorNotFound:
                return self.server.name

    @property
    def type_(self):
        """Store type (*private*, *public*, *archive*)."""
        table = self.server.sa.OpenUserStoresTable(MAPI_UNICODE)
        table.Restrict(SPropertyRestriction(RELOP_EQ, PR_EC_STOREGUID,
            SPropValue(PR_EC_STOREGUID, _bdec(self.guid))), TBL_BATCH)
        for row in table.QueryRows(1, 0):
            storetype = PpropFindProp(row, PR_EC_STORETYPE)
            if storetype:
                return {
                    ECSTORE_TYPE_PRIVATE: 'private',
                    ECSTORE_TYPE_ARCHIVE: 'archive',
                    ECSTORE_TYPE_PUBLIC: 'public',
                }[storetype.Value]

    @property
    def freebusy(self):
        """:class:`Freebusy <Freebusy>` information."""
        return FreeBusy(self)

    def _name_id(self, name_tuple):
        # TODO: use _pidlid_proptag everywhere
        id_ = self._name_id_cache.get(name_tuple)
        if id_ is None:
            named_props = [MAPINAMEID(*name_tuple)]
            # TODO use MAPI_CREATE, or too dangerous because of
            # potential db overflow?
            id_ = self.mapiobj.GetIDsFromNames(named_props, 0)[0]
            self._name_id_cache[name_tuple] = id_
        return id_

    def _pidlid_proptag(self, pidlid):
        id_ = self._pidlid_cache.get(pidlid)
        if id_ is None:
            id_ = _name_to_proptag(pidlid, self.mapiobj)[0]
            self._pidlid_cache[pidlid] = id_
        return id_

    def __eq__(self, s): # TODO check same server?
        if isinstance(s, Store):
            return self.guid == s.guid
        return False

    def subscribe(self, sink, **kwargs):
        """Subscribe to store notifications

        :param sink: Sink instance with callbacks to process notifications
        :param object_types: Tracked objects (*item*, *folder*)
        :param folder_types: Tracked folders (*mail*, *contacts*, *calendar*)
        :param event_types: Event types (*created*, *updated*, *deleted*)
        """
        _notification.subscribe(self, None, sink, **kwargs)

    def unsubscribe(self, sink):
        _notification.unsubscribe(self, sink)

    def __ne__(self, s):
        return not self == s

    def __iter__(self):
        return self.folders()

    def __unicode__(self):
        return "Store('%s')" % self.guid
